from os import path

import pandas as pd

pd.set_option('display.max_columns', 100)


class CriteoV2:

    def __init__(self, path_folder: str):
        # Define paths
        self.criteo_v2_path_original = path_folder + "criteo-uplift-v2.1.csv"
        self.criteo_v2_path = path_folder + "criteo_uplift_v2.csv"
        self.criteo_v2_path_resampled = path_folder + "criteo_uplift_v2_resampled.csv"

    def prep(self, resample: bool=False):
        """
        Prepare the Criteo_V2 dataset and store the csv files in the filesystem.

        1. Rename columns
        2. Downsample majority class (treatment group)
        3. Delete unnecessary columns (visit, exposure)
        4. Drop duplicates

        :param resample: True if the dataset should be downsampled. False otherwise. Default: False.
        """

        if resample:
            if path.exists(self.criteo_v2_path):
                return pd.read_csv(self.criteo_v2_path)
        else:
            if path.exists(self.criteo_v2_path_resampled):
                return pd.read_csv(self.criteo_v2_path_resampled)

        data = pd.read_csv(self.criteo_v2_path_original)

        # 1. Rename "conversion" column to "response"
        data.rename(columns={
            "conversion": "response"
            }, inplace=True)

        # 2. Downsample the treatment group
        if resample:
            data_no_treatment = data.loc[data.treatment == 0]
            data_treatment = data.loc[data.treatment == 1]

            frac = round(data_no_treatment.shape[0] / data_treatment.shape[0], 3)
            data_treatment = data_treatment.sample(frac=frac)
            data = pd.concat([data_treatment, data_no_treatment]).sample(frac=1)

        # 3. Delete unnecessary columns
        data.drop(["visit", "exposure"], axis=1, inplace=True)

        # 4. Remove duplicates
        data.drop_duplicates(inplace=True, ignore_index=True)
        data.reset_index(inplace=True, drop=True)

        if resample:
            data.to_csv(self.criteo_v2_path_resampled, index=False)
        else:
            data.to_csv(self.criteo_v2_path, index=False)

        return data


import logging
import pickle
from datetime import datetime

import pandas as pd
from sklearn.linear_model import LogisticRegression

from autoum.approaches.utils import ApproachParameters, DataSetsHelper, Helper


class TreatmentDummy:
    """
    Treatment dummy approach with Logistic Regression proposed by Lo (2002)
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for the treatment dummy approach with logistic regression

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """

        self.parameters = parameters
        self.cost_sensitive = approach_parameters.cost_sensitive
        self.feature_importance = approach_parameters.feature_importance
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using the treatment dummy approach.

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        df_train_with_interactions = data_set_helper.df_train.copy()
        df_train_with_interactions.drop(['response'], axis=1, inplace=True)
        if data_set_helper.valid:
            df_valid_with_interactions = data_set_helper.df_valid.copy()
            df_valid_with_interactions.drop(['response'], axis=1, inplace=True)
        df_test_with_interactions = data_set_helper.df_test.copy()
        df_test_with_interactions.drop(['response'], axis=1, inplace=True)

        columns = df_test_with_interactions.drop(['treatment'], axis=1).columns

        # Calculate interaction effects
        for column in columns:
            df_train_with_interactions[column + "_x_treatment"] = df_train_with_interactions[column] * df_train_with_interactions["treatment"]
            df_test_with_interactions[column + "_x_treatment"] = df_test_with_interactions[column] * df_test_with_interactions["treatment"]

            if data_set_helper.valid:
                df_valid_with_interactions[column + "_x_treatment"] = df_valid_with_interactions[column] * df_valid_with_interactions["treatment"]

        if self.cost_sensitive:
            # Calculate class weights
            class_weights = Helper.create_class_weight(data_set_helper.y_train)
            clf = LogisticRegression(class_weight=class_weights)
        else:
            clf = LogisticRegression()

        clf.set_params(**self.parameters)
        self.log.debug("Start fitting Treatment Dummy ...")

        clf.fit(df_train_with_interactions.to_numpy(), data_set_helper.y_train)
        self.log.debug(clf)
        response_dict = {}

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_Treatment_Dummy_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(clf, open(filename, 'wb'))

        self.log.debug("Predicting ...")

        # Training
        response_dict["score_train"] = TreatmentDummy.prediction(clf, df_train_with_interactions)

        # Validation
        if data_set_helper.valid:
            response_dict["score_valid"] = TreatmentDummy.prediction(clf, df_valid_with_interactions)

        # Test
        response_dict["score_test"] = TreatmentDummy.prediction(clf, df_test_with_interactions)

        return response_dict

    @staticmethod
    def prediction(clf, df_with_interactions: pd.DataFrame):
        """
        Predict the uplift scores for the given data set

        :param clf: Classifier
        :param df_with_interactions: DataFrame which should be analyzed
        :return: Uplift scores
        """

        df_with_interactions["treatment"] = 1
        score_train_t_1 = clf.predict_proba(df_with_interactions.to_numpy())[:, 1]
        df_with_interactions["treatment"] = 0
        score_train_t_0 = clf.predict_proba(df_with_interactions.to_numpy())[:, 1]

        return score_train_t_1 - score_train_t_0



from os import path

import pandas as pd
from sklearn import preprocessing

pd.set_option('display.max_columns', 100)


class Lenta:

    def __init__(self, path_folder: str):
        # Define paths

        self.lena_path_original = path_folder + "lenta_dataset.csv"
        self.lenta_path = path_folder + "lenta-dataset.csv"

    def prep(self):
        """
        Prepare the Lenta dataset and store the csv files in the filesystem.

        1. Rename columns
        2. Delete columns with noo many missing values (as this data set has a lot of missing values)
        3. Remove rows containing any missing values

        Note that we reduce the number of rows from 687.029 to 176.065. Further, we reduce the number of columns from 195 to 110
        :return:
        """

        if path.exists(self.lenta_path):
            return pd.read_csv(self.lenta_path)

        data = pd.read_csv(self.lena_path_original)

        # 1. Rename "response_att" column to "response"
        data.rename(columns={
            "response_att": "response",
            "group": "treatment"
            }, inplace=True)

        # 2. Remove columns with too many missing values
        threshold = 0.8  # 80% of the values (for each column) should NOT be missing
        data.dropna(thresh=int(threshold * data.shape[0]), axis=1, inplace=True)

        # 3. Remove rows containin any missing values
        data.dropna(inplace=True)
        data.reset_index(inplace=True, drop=True)

        # 4. Transform gender variable
        le = preprocessing.LabelEncoder()
        data.gender = le.fit_transform(data.gender)

        # 4. Transform treatment variable
        data["treatment"] = [0 if x == "control" else 1 for x in data.treatment]

        data.to_csv(self.lenta_path, index=False)

        return data



import logging
import pickle
from datetime import datetime

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier

from autoum.approaches.utils import ApproachParameters, DataSetsHelper, Helper


class LaisGeneralization:
    """
    Lais Generalization proposed by Kane et al. (2014)

    A new target variable is created using an arbitrary transformation.

    Assumption (without Generalization):
    P(T=1) = P(T=1) = 1/2 !!

    Thanks to Kane (2015) we can cope with any P(T) and P(C) this is often refered to as "Lais Generalization"
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for Lai's (Lais 2015) generalization (Kane et al, 2015)

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """
        self.parameters = parameters
        self.cost_sensitive = approach_parameters.cost_sensitive
        self.feature_importance = approach_parameters.feature_importance
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using Lai's generalization

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        # Transform the response
        y_train = LaisGeneralization.transform(data_set_helper.df_train)

        unique_classes = np.unique(y_train)
        if len(unique_classes) < 4:
            raise Exception("Training set only contains samples from three or less class. Available Classses {}".format(unique_classes))

        if self.cost_sensitive:
            # Calculate class weights
            class_weights = Helper.create_class_weight(y_train)
            clf = RandomForestClassifier(class_weight=class_weights)
        else:
            clf = RandomForestClassifier()

        clf.set_params(**self.parameters)

        self.log.debug("Start fitting Lai's Generalization ...")

        clf.fit(data_set_helper.x_train, y_train)

        self.log.debug(clf)
        transformed_dict = {}

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_LG_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(clf, open(filename, 'wb'))

        if self.feature_importance:
            transformed_dict["feature_importance"] = clf.feature_importances_

        self.log.debug("Predicting ... ")

        df_train = data_set_helper.df_train
        df_test = data_set_helper.df_test

        prob_t_train = df_train.loc[df_train.treatment == 1].shape[0] / df_train.shape[0]
        prob_t_test = df_test.loc[df_test.treatment == 1].shape[0] / df_test.shape[0]

        x_train = data_set_helper.x_train
        x_test = data_set_helper.x_test

        # (P(TR) / P(T)) + (P(CN) / P(C)) - (P(TN) / P(T)) - (P(CR) / P(C))
        transformed_dict["score_train"] = (clf.predict_proba(x_train)[:, 3] / prob_t_train) + (clf.predict_proba(x_train)[:, 0] / 1 - prob_t_train) - (
                clf.predict_proba(x_train)[:, 1] / prob_t_train) - (clf.predict_proba(x_train)[:, 2] / 1 - prob_t_train)
        transformed_dict["score_test"] = (clf.predict_proba(x_test)[:, 3] / prob_t_test) + (clf.predict_proba(x_test)[:, 0] / 1 - prob_t_test) - (
                clf.predict_proba(x_test)[:, 1] / prob_t_test) - (clf.predict_proba(x_test)[:, 2] / 1 - prob_t_test)

        if data_set_helper.valid:
            df_valid = data_set_helper.df_valid
            prob_t_valid = df_valid.loc[df_valid.treatment == 1].shape[0] / df_valid.shape[0]
            x_valid = data_set_helper.x_valid
            transformed_dict["score_valid"] = (clf.predict_proba(x_valid)[:, 3] / prob_t_valid) + (clf.predict_proba(x_valid)[:, 0] / 1 - prob_t_valid) - (
                    clf.predict_proba(x_valid)[:, 1] / prob_t_valid) - (clf.predict_proba(x_valid)[:, 2] / 1 - prob_t_valid)
        else:
            transformed_dict["score_valid"] = []

        return transformed_dict

    @staticmethod
    def transform(_df_train: pd.DataFrame) -> np.ndarray:
        """
        Transforms the target variable y with the following transformation

        T = 0 & R = 0 --> 1
        T = 1 & R = 0 --> 2
        T = 0 & R = 1 --> 3
        T = 1 & R = 1 --> 4

        Assumption: During the campaign, the treatment was randomly assigned to individuals.

        :param _df_train: Trainings DataFrame
        :return: Transformed target variable
        """

        df = _df_train.copy()
        df.loc[((df['treatment'] == 0) & (df['response'] == 0)), 'group'] = 1
        df.loc[((df['treatment'] == 1) & (df['response'] == 0)), 'group'] = 2
        df.loc[((df['treatment'] == 0) & (df['response'] == 1)), 'group'] = 3
        df.loc[((df['treatment'] == 1) & (df['response'] == 1)), 'group'] = 4

        return df["group"].to_numpy()



import logging
import pickle
from datetime import datetime

import numpy as np
from sklearn.ensemble import RandomForestClassifier

from autoum.approaches.utils import ApproachParameters, DataSetsHelper, Helper


class ClassVariableTransformation:
    """
    Class Variable Transformation proposed by Jaskowski & Jaroszewicz (2012)

    A new target variable is created using the following transformation:
    response * treatment + (1 - response) * (1 - treatment)

    Assumption:
    P(T=0) = P(T=1) = 1/2 !!

    The classifier is based on Random Forest algorithm.
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for the class variable transformation (Jaskowski & Jaroszewicz, 2012))

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """
        self.parameters = parameters
        self.cost_sensitive = approach_parameters.cost_sensitive
        self.feature_importance = approach_parameters.feature_importance
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using class variable transformation

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        # Sanity Check: Check if treatment and control group are almost equal (maximum difference being 0.1)
        df_treatment = data_set_helper.df_train['treatment']
        diff = np.abs(df_treatment.value_counts()[1] / df_treatment.shape[0] - df_treatment.value_counts()[0] / df_treatment.shape[0])
        if diff >= 0.1:
            self.log.error(
                f"Assumption P(G=T)=P(G=C)=1/2 for the transformed outcome approach by Jaskowski & Jaroszewicz (2012) is violated with a difference between treatment and"
                f"control group of {diff:.3f}. This approach will be skipped!")
            raise ValueError("Assumption P(G=T)=P(G=C)=1/2 for the transformed outcome approach by Jaskowski & Jaroszewicz (2012) is violated")

        # Transform the response
        y_train = ClassVariableTransformation.transform(data_set_helper.df_train['response'].to_numpy(), data_set_helper.df_train['treatment'].to_numpy())
        unique_classes = np.unique(y_train)
        if len(unique_classes) < 2:
            raise Exception("Training set only contains samples from one class. Availabe Class : {}".format(unique_classes))

        if self.cost_sensitive:
            # Calculate class weights
            class_weights = Helper.create_class_weight(y_train)
            clf = RandomForestClassifier(class_weight=class_weights)
        else:
            clf = RandomForestClassifier()

        clf.set_params(**self.parameters)

        self.log.debug("Start fitting Class Variable Transformation ...")

        clf.fit(data_set_helper.x_train, y_train)
        transformed_dict = {}

        if self.feature_importance:
            transformed_dict["feature_importance"] = clf.feature_importances_

        self.log.debug(clf)

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_CVT_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(clf, open(filename, 'wb'))

        self.log.debug("Predicting ... ")

        transformed_dict["score_train"] = 2 * clf.predict_proba(data_set_helper.x_train)[:, 1] - 1
        transformed_dict["score_test"] = 2 * clf.predict_proba(data_set_helper.x_test)[:, 1] - 1
        if data_set_helper.valid:
            transformed_dict["score_valid"] = 2 * clf.predict_proba(data_set_helper.x_valid)[:, 1] - 1
        else:
            transformed_dict["score_valid"] = []

        return transformed_dict

    @staticmethod
    def transform(y: np.ndarray, treat: np.ndarray) -> np.ndarray:
        """
        Transforms the target variable y with the following transformation

        y_trans = y * treat + (1-y) * (1-treat)

        Assumption: During the campaign, the treatment was randomly assigned to individuals.

        :param y: Target variable to be transformed
        :param treat: Treatment variable
        :return: Transformed target variable
        """

        return y * treat + (1 - y) * (1 - treat)



import pandas as pd
from sklearn import preprocessing
from os import path

pd.set_option('display.max_columns', 100)


class Starbucks:

    def __init__(self, path_folder: str):
        # Define paths
        self.starbucks_train_path = path_folder + "Training.csv"
        self.starbucks_test_path = path_folder + "Test.csv"
        self.starbucks_path = path_folder + "starbucks.csv"

    def prep(self):
        """
        1. Rename columns
        2. Encode treatment column
        3. Remove duplicates
        """
        if path.exists(self.starbucks_path):
            return pd.read_csv(self.starbucks_path)

        data_train = pd.read_csv(self.starbucks_train_path)
        data_test = pd.read_csv(self.starbucks_test_path)

        # 1. Rename "conversion" column to "response"
        data_train.rename(columns={
            "purchase": "response",
            "Promotion": "treatment"
            }, inplace=True)
        data_test.rename(columns={
            "purchase": "response",
            "Promotion": "treatment"
            }, inplace=True)

        # 2. Encode treatment column
        le = preprocessing.LabelEncoder()
        data_train.treatment = le.fit_transform(data_train.treatment)
        data_test.treatment = le.fit_transform(data_test.treatment)

        data = pd.concat([data_train, data_test]).sample(frac=1)
        data.drop(["ID"], axis=1, inplace=True)

        le = preprocessing.LabelEncoder()
        data["V1"] = le.fit_transform(data["V1"])
        data["V5"] = le.fit_transform(data["V5"])
        data["V6"] = le.fit_transform(data["V6"])

        # 3. Remove duplicates
        data.drop_duplicates(inplace=True, ignore_index=True)
        data.reset_index(inplace=True, drop=True)

        data.to_csv(self.starbucks_path, index=False)

        return data



import logging
import os
from datetime import datetime
from decimal import Decimal, ROUND_HALF_UP

import numpy as np
import pandas as pd
from deprecated import deprecated
from matplotlib import pyplot as plt

from autoum.const.const import *

logging = logging.getLogger(__name__)


class UpliftEvaluation:
    """
    Class for evaluating uplift performance
    """

    @staticmethod
    def separate_treated_control(df_results: pd.DataFrame, treatment_col: str, response_col: str, uplift_score_col: str, treated_asc: bool = False, control_asc: bool = False):
        """
        Separate the given dataframe into two dataframes. One containing the treated samples and the other one containing the control samples.

        :param df_results: DataFrame containing treated and control samples
        :param treatment_col: Name of the treatment column
        :param response_col: Name of the response column
        :param uplift_score_col: Name of the uplift column
        :param treated_asc: True, if the treated DataFrame should be sorted by ascending order. False otherwise. Default: False
        :param control_asc: True, if the control DataFrame should be sorted by ascending order. False otherwise. Default: False
        :return: Tuple df_results_treated, df_results_control
        """
        df_results_treated = df_results.loc[df_results[treatment_col] == 1].copy()
        df_results_control = df_results.loc[df_results[treatment_col] == 0].copy()

        df_results_treated.sort_values(by=[uplift_score_col, response_col], ascending=[treated_asc, treated_asc], inplace=True)
        df_results_treated.reset_index(drop=True, inplace=True)
        df_results_control.sort_values(by=[uplift_score_col, response_col], ascending=[control_asc, control_asc], inplace=True)
        df_results_control.reset_index(drop=True, inplace=True)

        return df_results_treated, df_results_control

    @staticmethod
    def calculate_uplift(bin_treatment_responder: list,
                         bin_number_treated_samples: list,
                         bin_non_treatment_responder: list,
                         bin_number_non_treated_samples: list,
                         bin_list: list,
                         num_treated_samples):
        """
        Given the treatment responder, treatment non responder as well as the number of treated samples and number of non-treated samples for **each** bin, calculate the
        uplift score for **each** bin using Radcliffe's defintion of uplift: r_ts - ((r_cs * n_ts) / n_cs)

        :param bin_treatment_responder: List which contains the number of treatment responders for each bin
        :param bin_number_treated_samples: List which contains the number of treatment samples for each bin
        :param bin_non_treatment_responder: List which contains the number of control (non-treatment) responders for each bin
        :param bin_number_non_treated_samples: List which contains the number of control (non-treatment) responders for each bin
        :param bin_list: List with the bin values (percentiles)
        :param num_treated_samples: Number of treatment samples (overall)
        :return: Tuple bins, uplift scores (absolute) and uplift scores (relative)
        """

        bin_treatment_responder = np.array(bin_treatment_responder)
        bin_number_treated_samples = np.array(bin_number_treated_samples)
        bin_non_treatment_responder = np.array(bin_non_treatment_responder)
        bin_number_non_treated_samples = np.array(bin_number_non_treated_samples)
        bin_list = np.array(bin_list)

        # Calculate uplift
        uplift = np.around(bin_treatment_responder - np.divide(bin_non_treatment_responder * bin_number_treated_samples, bin_number_non_treated_samples,
                                                               out=np.zeros(bin_number_non_treated_samples.shape, dtype=float), where=bin_number_non_treated_samples != 0), 4)

        # Add starting point (0,0)
        bin_list = np.insert(bin_list, 0, 0)
        uplift = np.insert(uplift, 0, 0)
        uplift_pct = np.around(uplift / num_treated_samples, 4)

        return bin_list, uplift, uplift_pct

    @staticmethod
    def calculate_qini_curve(df_results: pd.DataFrame, uplift_score_col: str, bins: int = 10, treatment_col: str = 'treatment', response_col: str = 'response'):
        """
        Calculate the qini curve, once with absolute uplift values for each bin, and once with relative uplift values for each bin using Radcliffe's defintion of uplift:
        r_ts - ((r_cs * n_ts) / n_cs)

        The data set (including treatment, response and score column) is divided into treatment and control group. Both tables are sorted in descending order according to their
        score column. In order to calculate the uplift value for the first segment (e.g. decile with 10%), we have to use the first segment in the treatment and the first segment
        in the control group. Example: total_number_of_samples = 20, treated_samples = 14 and control_samples = 6 . 10% of treated_samples = 1 (rounded down) and 10% of
        control_samples = 1 (rounded up). Consequently, the calculations for the uplift value are based on one sample from the treatment group and one sample from the
        control group. Therefore, the calculations refer to the segments of treatment and control tables, respectively!

        :param df_results: DataFrame containing treated and control samples
        :param uplift_score_col: Name of the uplift column
        :param bins: Number of bins. Default: 10 (deciles)
        :param treatment_col: Name of the treatment column. Default: 'treatment'
        :param response_col: Name of the response column. Default: 'response'
        :return: Tuple bins, uplift scores (absolute) and uplift scores (relative)
        """

        df_results_treated, df_results_control = UpliftEvaluation.separate_treated_control(df_results, treatment_col, response_col, uplift_score_col)

        bin_list, r_ts, n_ts, r_cs, n_cs = UpliftEvaluation.calculate_group_share(df_results_treated, df_results_control, bins, response_col)

        return UpliftEvaluation.calculate_uplift(r_ts, n_ts, r_cs, n_cs, bin_list, df_results_treated.shape[0])

    @staticmethod
    def calculate_optimal_qini_curve(df_results: pd.DataFrame, bins: int = 10, treatment_col: str = 'treatment', response_col: str = 'response'):
        """
        Calculate optimal qini curve including negative effects. The following applies for an optimal qini curve:

        Treatment Responders > Treatment Non Responders and
        Control Non Responders > Control Responders

        :param df_results: DataFrame containing treated and control samples
        :param bins: Number of bins. Default: 10 (deciles)
        :param treatment_col: Name of the treatment column. Default: 'treatment'
        :param response_col: Name of the response column. Default: 'response'
        :return: Tuple bins, uplift scores (absolute) and uplift scores (relative)
        """

        df_results.loc[((df_results[treatment_col] == 0) & (df_results[response_col] == 0)), 'group'] = 2
        df_results.loc[((df_results[treatment_col] == 1) & (df_results[response_col] == 0)), 'group'] = 1
        df_results.loc[((df_results[treatment_col] == 0) & (df_results[response_col] == 1)), 'group'] = 0
        df_results.loc[((df_results[treatment_col] == 1) & (df_results[response_col] == 1)), 'group'] = 3

        df_results_treated = df_results.loc[df_results[treatment_col] == 1].copy()
        df_results_control = df_results.loc[df_results[treatment_col] == 0].copy()

        df_results_treated.sort_values(by=['group'], ascending=[False], inplace=True)
        df_results_treated.reset_index(drop=True, inplace=True)
        df_results_treated.drop(['group'], axis=1, inplace=True)

        df_results_control.sort_values(by=['group'], ascending=[False], inplace=True)
        df_results_control.reset_index(drop=True, inplace=True)
        df_results_control.drop(['group'], axis=1, inplace=True)

        bin_list, r_ts, n_ts, r_cs, n_cs = UpliftEvaluation.calculate_group_share(df_results_treated, df_results_control, bins, response_col)

        return UpliftEvaluation.calculate_uplift(r_ts, n_ts, r_cs, n_cs, bin_list, df_results.loc[df_results[treatment_col] == 1].shape[0])

    @staticmethod
    def calculate_random_qini_curve(bins: int, endpoint_abs: int, endpoint_pct: int, increase_value: float = 0.0):
        """
        Calculate a random qini curve

        :param bins: Number of bins
        :param endpoint_abs: Endpoint of the random curve using absolute values
        :param endpoint_pct: Endpoint of the random curve using relative values
        :param increase_value: If the endpoint is negative, we will increase each value by increase_value (because of integration issues with negative areas)
        :return: Tuple bins, uplift scores (absolute) and uplift scores (relative)
        """

        uplift_abs = np.around(np.linspace(0.0 + increase_value, endpoint_abs + increase_value, num=bins + 1), 4)
        uplift_pct = np.around(np.linspace(0.0 + increase_value, endpoint_pct + increase_value, num=bins + 1), 4)
        bin_list = np.linspace(0, 1, num=bins + 1)

        return bin_list, uplift_abs, uplift_pct

    @staticmethod
    def calculate_group_share(df_treated: pd.DataFrame, df_control: pd.DataFrame, bins: int = 10, response_col: str = 'response'):
        """
        Calculate the share of groups inherent in any data set: Treatment responder (r_ts), Treatment non-respondes (n_ts), Control responder (r_cs), Control non-responder (n_cs)

        :param df_treated: DataFrame containg the results for the treatment group
        :param df_control: DataFrame containing the results for the control group
        :param bins: Number of bins. Default: 10 (deciles)
        :param response_col: Name of the response column. Default: 'response'
        :return: bin_list (list of bins), r_ts (treatment responder), n_ts (treatment non responder), r_cs (control responder), n_cs (control non responder)
        """
        bin_list = []
        r_ts = []
        n_ts = []
        r_cs = []
        n_cs = []

        for bin_idx in np.linspace(0, 1, bins + 1):
            if bin_idx == 0.0:
                continue
            bin_idx = int(bin_idx * 1000) / float(1000)
            bin_list.append(bin_idx)
            number_of_treated_samples_bin = int(Decimal(bin_idx * df_treated.shape[0]).to_integral_value(rounding=ROUND_HALF_UP))
            number_of_control_samples_bin = int(Decimal(bin_idx * df_control.shape[0]).to_integral_value(rounding=ROUND_HALF_UP))

            if number_of_treated_samples_bin > 0 and number_of_control_samples_bin > 0:
                r_ts.append(df_treated.iloc[0:number_of_treated_samples_bin].loc[(df_treated[response_col] == 1)].shape[0])
                n_ts.append(number_of_treated_samples_bin)
                r_cs.append(df_control.iloc[0:number_of_control_samples_bin].loc[(df_control[response_col] == 1)].shape[0])
                n_cs.append(number_of_control_samples_bin)
            else:
                r_ts.append(0)
                n_ts.append(0)
                r_cs.append(0)
                n_cs.append(0)

        return bin_list, r_ts, n_ts, r_cs, n_cs

    @staticmethod
    def calculate_actual_uplift_in_bins(df_results: pd.DataFrame, bins: int = 10, treatment_col: str = 'treatment', response_col: str = 'response', absolute: bool = False) -> dict:
        """
        Calculate the uplift value for each decile and uplift modeling approach

        The uplift in each decile is defined as treatment response rate - control response rate

        :param df_results: DataFrame containing treated and control samples
        :param bins: Number of bins. Default: 10 (deciles)
        :param treatment_col: Name of the treatment column. Default: 'treatment'
        :param response_col: Name of the response column. Default: 'response'
        :param absolute: If True, return absolute values, relative values otherwise. Default: False
        :return: Dictionary containing the uplift values for each decile and approach
        """

        uplift_in_deciles = {}

        for col in df_results.drop(["response", "treatment"], axis=1).columns:
            # If the uplift modeling approach runs into an error and returns nan values; in such a case, ignore the evaluation
            if df_results[col].isnull().all():
                continue
            try:
                bin_list, uplift_abs, uplift_pct = UpliftEvaluation.calculate_qini_curve(df_results, col, bins, treatment_col, response_col)

                if absolute:
                    uplift_in_deciles.update(UpliftEvaluation.store_uplift_in_bins(uplift_abs, col + '-Uplift_'))
                else:
                    uplift_in_deciles.update(UpliftEvaluation.store_uplift_in_bins(uplift_pct, col + '-Uplift_'))

            except Exception as e:
                print("Error in calculate_uplift_in_deciles")
                print(e)
                uplift_in_deciles[col + '-Uplift_'] = np.nan

        return uplift_in_deciles

    @staticmethod
    def calculate_optimal_uplift_in_bins(df_results: pd.DataFrame, bins: int = 10, treatment_col: str = 'treatment', response_col: str = 'response', absolute: bool = False):
        """
        Calculate the optimal uplift value for each decile

        :param df_results: DataFrame (fold) which contains the treatment and control samples inclduing their responses
        :param bins: Number of bins. Default: 10 (deciles)
        :param treatment_col: Name of the treatment column. Default: 'treatment'
        :param response_col: Name of the response column. Default: 'response'
        :param absolute: If True, return absolute values, relative values otherwise. Default: False
        :return: Dictionary containing the optimal uplift values for each decile and approach
        """

        _, uplift_abs, uplift_pct = UpliftEvaluation.calculate_optimal_qini_curve(df_results, bins=bins, treatment_col=treatment_col, response_col=response_col)

        if absolute:
            return UpliftEvaluation.store_uplift_in_bins(uplift_abs, 'Optimal-Uplift_')
        else:
            return UpliftEvaluation.store_uplift_in_bins(uplift_pct, 'Optimal-Uplift_')

    @staticmethod
    def store_uplift_in_bins(uplift: np.ndarray, col_name: str):
        """
        Stores the given uplift values in a dictionary containing the column name and bin number as key and the uplift value as vaue

        :param uplift: List of uplift values
        :param col_name: Name of the key
        :return: Dictionary containing the uplift values for each bin
        """
        uplift_in_deciles = {}

        for i, decile in enumerate(uplift):
            uplift_in_deciles[col_name + str(i)] = decile

        return uplift_in_deciles

    @staticmethod
    def calculate_unscaled_qini_coefficient(df_uplift_bins: pd.DataFrame, bins: int = 10):
        """
        Calculate the unscaled qini coefficient for each approach and split (if k-fold CV):

        The unscaled qini coefficient is defined as the ratio of the actual uplift gains curve above the diagonal. Optionally divded by N^2

        The range is between 0 and infinity. The higher the score, the better the model

        Example:

        Approach A has an integral of 10, Approach B has an integral of 20, and Random Targeting has an integral of 5.
        Thus, Approach A is twice as good as Random Targeting (10/5=2), and Approach B is four times as good as Random Targeting (20/5=4).
        Thus, we can argue that the UQC expresses how much better an Approach is over Random Targeting

        :param df_uplift_bins: DataFrame containing the uplift values for each bin, approach, and fold
        :param bins: Number of bins. Default 10 (Deciles)
        :return: DataFrame which contains the uplift values (for each bin) and the unscaled qini coefficient for each approach and split
        """

        df_metrics = pd.DataFrame()

        for approach_idx in range(bins + 1, df_uplift_bins.shape[1] + 1, bins + 1):  # for each appraoch
            # Get name of uplift modeling approach
            approach_name = df_uplift_bins.columns.values[approach_idx - 11].split("-")[0]

            # Get uplift values in deciles (bins)
            df_uplift_bin_approach = df_uplift_bins.iloc[:, -(bins + 1) + approach_idx:approach_idx].copy()

            # Calculate UQC for a given split and approach
            uqc_list = []
            for row_idx in range(0, df_uplift_bins.shape[0]):  # for each split
                uplift_bins = df_uplift_bins.iloc[row_idx, -(bins + 1) + approach_idx:approach_idx].values

                endpoint = uplift_bins[-1]

                # Case 1: Endpoint is positive
                if endpoint > 0:
                    # Calculate area under random curve
                    bin_list_rand, uplift_ran, _ = UpliftEvaluation.calculate_random_qini_curve(bins, endpoint, endpoint)

                # Case 2: Endpoint is negative
                else:
                    # Calculate area under random curve
                    bin_list_rand, uplift_ran, _ = UpliftEvaluation.calculate_random_qini_curve(bins, endpoint, endpoint, abs(endpoint))

                    # Increase each value by endpoint (because of negative and positive issues when calculating an integral)
                    uplift_bins = uplift_bins + np.abs(endpoint)

                area_under_random = np.trapz(uplift_ran, x=bin_list_rand)
                area_under_curve = np.trapz(uplift_bins, x=bin_list_rand)

                uqc_list.append(round(area_under_curve / area_under_random, 4))

            df_uplift_bin_approach[approach_name + "-UQC"] = uqc_list
            df_metrics = pd.concat([df_metrics, df_uplift_bin_approach], axis=1)

        return df_metrics

    @staticmethod
    def calculate_qini_coefficient(df_metrics: pd.DataFrame, df_opt_uplift_bins: pd.DataFrame, bins: int = 10, num_columns: int = 12) -> pd.DataFrame:
        """
        Calculate the qini coefficient for each approach and split (if k-fold CV):

        The qini coefficient is defined as the ratio of the actual uplift curve above the diagonal to that of the optimum qini curve.
        Here optimal does take into account negative effects.

        The range is between -1 and 1. The higher the score, the better the model

        :param df_metrics: DataFrame containing the uplift values (for each bin) and (optionally) the unscaled qini coefficient for each approach and split
        :param df_opt_uplift_bins: DataFrame containing the optimal uplift values for each decile and split
        :param bins: Number of bins. Default 10 (Deciles)
        :param num_columns: Number of columns per approach: With UQC it's bins + 2. Without UQC it's bins +1.
        :return: DataFrame which contains the uplift values (for each bin), unscaled qini coefficient, and (optionally) the qini coefficient for each approach and split
        """

        if df_metrics.shape[0] != df_opt_uplift_bins.shape[0]:
            raise ValueError("Error in calculate_qini_coefficient. The number of rows of df_metrics and df_opt_uplift_bins is unequal")

        # Final DataFrame containing the uplift values (for each bin), unscaled qini coefficient, and the qini coefficient for each approach and split
        df_metrics_qc_extended = pd.DataFrame()

        for row_idx in range(0, df_metrics.shape[0]):  # for each split/row

            # 1. Calculate qini coefficients' denominator (globally; independent from approach)
            endpoint = df_opt_uplift_bins.iloc[row_idx].values[-1]
            uplift_opt = df_opt_uplift_bins.iloc[row_idx].values

            # Case 1: Endpoint is positive
            if endpoint > 0:
                # Calculate random curve
                bin_list_rand, uplift_rand, _ = UpliftEvaluation.calculate_random_qini_curve(bins, endpoint, endpoint)

            # Case 2: Endpoint is negative
            else:
                # Calculate random curve
                bin_list_rand, uplift_rand, _ = UpliftEvaluation.calculate_random_qini_curve(bins, endpoint, endpoint, abs(endpoint))
                uplift_opt = uplift_opt + np.abs(endpoint)

            area_under_opt = np.trapz(uplift_opt, x=bin_list_rand)
            area_under_random = np.trapz(uplift_rand, x=bin_list_rand)

            # Calculate qini coefficients' demoninator
            qini_coefficient_denominator = area_under_opt - area_under_random

            df_metrics_row = pd.DataFrame()

            # 2. Calculate qini coefficients' numerator for each approach
            for approach_idx in range(num_columns, df_metrics.shape[1] + 1, num_columns):  # for each approach
                # Get name of modeling approach
                approach_name = df_metrics.columns.values[approach_idx - num_columns].split("-")[0]

                # Depending on whether UQC values have been calculated prior to the QC values, we need to ignore these values
                if approach_idx % 12 == 0:
                    uplift_bins_split = df_metrics.iloc[row_idx, -num_columns + approach_idx:approach_idx - 1].values
                else:
                    uplift_bins_split = df_metrics.iloc[row_idx, -num_columns + approach_idx:approach_idx].values

                if endpoint < 0:
                    # Increase each value by endpoint (because of issues when calculating the integral of negative and positive areas)
                    uplift_bins_split = uplift_bins_split + np.abs(endpoint)

                area_under_curve = np.trapz(uplift_bins_split, x=bin_list_rand)

                # Calculate qini coefficient
                qini_coefficient_numerater = area_under_curve - area_under_random

                df_metrics_approach = df_metrics.iloc[[row_idx], -num_columns + approach_idx:approach_idx]
                df_metrics_approach[approach_name + "-QC"] = round(qini_coefficient_numerater / qini_coefficient_denominator, 4)

                df_metrics_row = pd.concat([df_metrics_row, df_metrics_approach], axis=1)

            df_metrics_qc_extended = pd.concat([df_metrics_qc_extended, df_metrics_row], axis=0)

        return df_metrics_qc_extended

    @staticmethod
    def calculate_mean(df_metrics: pd.DataFrame, bins: int = 10, qc: bool = False):
        """

        :param df_metrics: DataFrame containing the uplift values for each bin, approach, and fold
        :param bins: Number of bins. Default 10 (Deciles)
        :param qc: True, if the qini coefficient was calculated during the evaluation. False otherwise. Default: False
        :return: DataFrame which contains the geometric mean uplift values (for each bin), arithmetic mean unscaled qini coefficient, and (optionally) the arithmetic mean qini coefficient for each approach and split
        """
        df_metrics_mean = pd.DataFrame()

        num_cols = bins + 3 if qc else bins + 2

        for approach_idx in range(num_cols, df_metrics.shape[1] + 1, num_cols):  # for each approach

            # Get name of uplift modeling approach
            approach_name = df_metrics.columns.values[approach_idx - num_cols].split("-")[0]

            df_metrics_approach = df_metrics.iloc[:, -num_cols + approach_idx:approach_idx]

            # Calcuate arithmetic mean of uplift values for each approach and bin (decile) across the splits
            uplift_cols = [col for col in df_metrics_approach.columns if 'QC' not in col]
            arithmetic_uplift_bins = df_metrics_approach[uplift_cols].mean(axis=0).to_frame().transpose().values[0].round(4)

            # Calculate the arithmetic mean of unscaled qini coefficient values for each approach across the splits
            uqc_cols = [col for col in df_metrics_approach.columns if 'UQC' in col]
            arithmetic_uqc = np.mean(df_metrics_approach[uqc_cols].values).round(4)

            # Create a mean DataFrame
            mean_df = pd.DataFrame(arithmetic_uplift_bins).transpose()
            mean_df.columns = uplift_cols
            mean_df[approach_name + "-UQC"] = arithmetic_uqc
            # Optionally, calculate the arithmetic mean of qini coefficient values for each approach across the splits
            if qc:
                qc_cols = [col for col in df_metrics_approach.columns if '-QC' in col]
                arithmetic_qc = np.mean(df_metrics_approach[qc_cols].values).round(4)
                mean_df[approach_name + "-QC"] = arithmetic_qc

            df_metrics_mean = pd.concat([df_metrics_mean, mean_df], axis=1)

        return df_metrics_mean

    @staticmethod
    def plot_qini_curve(df_metrics: pd.DataFrame,
                        df_opt_uplift_bins: pd.DataFrame,
                        bins: int = 10,
                        plot_optimum: bool = False,
                        absolute: bool = False,
                        title: str = 'Qini Curve',
                        save_figure: bool = False,
                        path: str = None,
                        plot_grayscale: bool = False,
                        plot_uqc: bool = False,
                        show_title=False,
                        fontsize=14,
                        qc: bool = False):
        """
        Plot the qini curves for each uplift approach in the given DataFrame including random and (optionally) optimal qini curves.

        :param df_metrics: DataFrame containg the metrics such as unscaled qini coefficient or qini coefficient
        :param df_opt_uplift_bins: DataFrame containing the optimal uplift values for each bin and approach
        :param bins: Number of bins. Default: 10
        :param plot_optimum: True if the optimum curve should be plotted. False otherwise. Default: False
        :param absolute: True if the absolute uplift scores should be plotted. False otherwise. Default: False
        :param title: Title of the plot
        :param save_figure: True if the figure should be saved. False otherwise. Default: False
        :param path: Path where the figure should be saved
        :param plot_grayscale: True, if the graphics should be printed in grayscale. False otherwise. Default: False
        :param plot_uqc: True, if the legend should contain the values of the unscaled qini coefficient. False otherwise.
        :param show_title: True if the title should be shown. False otherwise. Default: False
        :param fontsize: Size of the elements in the graphic. Default: 14.
        :param qc: True, if the qini coefficient was calculated during the evaluation. False otherwise. Default: False
        """

        # Matplotlib Settings
        fig, axes = plt.subplots(figsize=(15, 10), facecolor="white")
        # Set facecolor of axes to white
        axes.set_facecolor('white')
        # Update fontsize
        plt.rcParams.update({
            'font.size': fontsize
        })
        plt.xticks(fontsize=fontsize)
        plt.yticks(fontsize=fontsize)

        ## Qini Curve
        bin_list = np.linspace(0, 1, bins + 1)
        endpoint = None

        num_cols = bins + 3 if qc else bins + 2
        for approach_idx in range(num_cols, df_metrics.shape[1] + 1, num_cols):  # for each approach plot one line

            approach_name = df_metrics.columns.values[approach_idx - num_cols].split("-")[0]
            approach_label = NAMING_SCHEMA[approach_name]
            df_metrics_approach = df_metrics.iloc[:, -num_cols + approach_idx:approach_idx]

            uplift_cols = [col for col in df_metrics_approach.columns if 'QC' not in col]
            uplift_values = df_metrics_approach[uplift_cols].values[0]

            if plot_uqc:
                uqc_cols = [col for col in df_metrics_approach.columns if 'UQC' in col]
                approach_label = approach_label + " - UQC: " + str(df_metrics_approach[uqc_cols].values[0][0])

            # Plot line
            if plot_grayscale:
                axes.plot(bin_list, uplift_values, label=approach_label, color=COLOR_SCHEMA_GRAY[approach_name], linestyle=LINESTYLES[approach_name])
            else:
                axes.plot(bin_list, uplift_values, label=approach_label, color=COLOR_SCHEMA[approach_name])

            # Endpoint
            if endpoint is None:
                endpoint = uplift_values[-1]

        # Plot random curve
        bin_list_rand, uplift_rand, _ = UpliftEvaluation.calculate_random_qini_curve(bins, endpoint, endpoint)
        axes.plot(bin_list_rand, uplift_rand, label='Random', color='black', linestyle='solid')

        # Plot optimal curve
        if plot_optimum:
            axes.plot(bin_list, df_opt_uplift_bins.mean(axis=0).to_frame().transpose().values[0].round(4), label='Optimum', color='cyan')

        if absolute:
            axes.set_ylabel("Cumulative Number of Incremental Responders / Uplift (Absolute)", fontsize=fontsize)
        else:
            axes.set_ylabel("Cumulative Number of Incremental Responders / Uplift (Relative)", fontsize=fontsize)

        axes.set_xlabel("Fraction of data targeted (in %)", fontsize=fontsize)

        # Sort both labels and handles by labels
        handles, labels = axes.get_legend_handles_labels()
        labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
        axes.legend(handles, labels, facecolor="white")

        plt.xticks([0.0, 0.2, 0.4, 0.6, 0.8, 1.0], ["0", "20", "40", "60", "80", "100"])

        if show_title:
            axes.set_title(title)

        if save_figure:
            if not os.path.exists(path):
                os.makedirs(path)
            now = datetime.now()
            now_date = now.date()
            plt.savefig(path + title.replace(" ", "_") + "_" + str(now_date) + ".png", facecolor=fig.get_facecolor(), transparent=True)

        plt.tight_layout()
        plt.show()

    @staticmethod
    def plot_gains_chart(df_results: pd.DataFrame):
        """
        Evaluate the results of the given dataset by calculating the cumulative sum. (Gains Chart)

        :param df_results: Pandas Dataframe. Each column is a score calculated by an algorithm e.g. Direct Uplift etc.
        """

        fig, axes = plt.subplots()

        # Random Line
        axes.plot((0, df_results.shape[0]), (0, df_results.respomse.sum()), label="Random selection", linestyle='dashed')

        for col in df_results.drop(["response", "treatment"], axis=1).columns:
            # Calculate cumulative sum for each column
            cumulative = np.cumsum(df_results.sort_values(by=col, ascending=False)["response"].tolist())
            # auc_score = auc(np.arange(0, 41650), cumulative)
            axes.plot(cumulative, label=col)

        axes.legend()
        axes.set_title("Gains Chart")
        plt.show()

    @staticmethod
    @deprecated(version='1.0', reason="This function is only implemented for the sake of completeness. Use calculate_qini_curve instead")
    def calculate_qini_curve2(df_results: pd.DataFrame, uplift_score_col: str, bins: int = 10, treatment_col: str = 'treatment', response_col: str = 'response'):
        """
        Calculate absolute and relative uplift values for each bin using Radcliffe's defintion of uplift: r_ts - ((r_cs * n_ts) / n_cs)

        The data set (including treatment, response and score column) is not split. The table is sorted in descending order according to the score column. The calculation for
        the uplift value of the first segment (e.g. decile with 10%) refers to the whole table. Example: total_number_of_samples = 20, treated_samples = 14, control_samples = 6.
        10% of the whole table = 2. Consequently, the calculations of the uplift values refer to the first two values of the table, regardless of whether they belong to the
        treatment or control group.

        :param df_results: DataFrame containing treated and control samples
        :param uplift_score_col: Name of the uplift column
        :param bins: Number of bins. Default: 10 (deciles)
        :param treatment_col: Name of the treatment column. Default: 'treatment'
        :param response_col: Name of the response column. Default: 'response'
        :return: Tuple bins, uplift scores (absolute) and uplift scores (relative)
        """

        df_results.sort_values(by=[uplift_score_col, response_col], ascending=[False, False], inplace=True)
        df_results.reset_index(drop=True, inplace=True)

        bin_list = []
        r_ts = []
        n_ts = []
        r_cs = []
        n_cs = []

        for bin_idx in np.linspace(0, 1, bins + 1):
            if bin_idx == 0.0:
                continue
            bin_idx = int(bin_idx * 1000) / float(1000)
            bin_list.append(bin_idx)
            number_of_samples_bin = int(Decimal(bin_idx * df_results.shape[0]).to_integral_value(rounding=ROUND_HALF_UP))

            if number_of_samples_bin > 0:
                df_results_treated = df_results.iloc[0:number_of_samples_bin].loc[df_results[treatment_col] == 1].copy()
                df_results_control = df_results.iloc[0:number_of_samples_bin].loc[df_results[treatment_col] == 0].copy()

                r_ts.append(df_results_treated.loc[(df_results_treated[response_col] == 1)].shape[0])
                n_ts.append(df_results_treated.shape[0])
                r_cs.append(df_results_control.loc[(df_results_control[response_col] == 1)].shape[0])
                n_cs.append(df_results_control.shape[0])
            else:
                r_ts.append(0)
                n_ts.append(0)
                r_cs.append(0)
                n_cs.append(0)

        return UpliftEvaluation.calculate_uplift(r_ts, n_ts, r_cs, n_cs, bin_list, df_results.loc[df_results[treatment_col] == 1].shape[0])

    @staticmethod
    @deprecated(version='1.0', reason="This function is only implemented for the sake of completeness. Use calculate_optimal_qini_curve instead")
    def calculate_optimal_qini_curve2(df_results: pd.DataFrame, bins: int, treatment_col: str, response_col: str):
        """
        Calculate optimal qini curve including negative effects. The following applies for an optimal qini curve:

        Treatment Responders > Treatment Non Responders > Control Non Responders > Control Responders

        :param df_results: DataFrame containing treated and control samples
        :param bins: Number of bins
        :param treatment_col: Name of the treatment column
        :param response_col: Name of the response column
        :return: Tuple bins, uplift scores (absolute) and uplift scores (relative)
        """

        df_results.loc[((df_results[treatment_col] == 0) & (df_results[response_col] == 0)), 'group'] = 2
        df_results.loc[((df_results[treatment_col] == 1) & (df_results[response_col] == 0)), 'group'] = 1
        df_results.loc[((df_results[treatment_col] == 0) & (df_results[response_col] == 1)), 'group'] = 0
        df_results.loc[((df_results[treatment_col] == 1) & (df_results[response_col] == 1)), 'group'] = 3

        df_results.sort_values(by=['group'], ascending=[False], inplace=True)
        df_results.reset_index(drop=True, inplace=True)
        df_results.drop(['group'], axis=1, inplace=True)

        bin_list = []
        r_ts = []
        n_ts = []
        r_cs = []
        n_cs = []

        for bin_idx in np.linspace(0, 1, bins + 1):
            if bin_idx == 0.0:
                continue
            bin_idx = int(bin_idx * 1000) / float(1000)
            bin_list.append(bin_idx)
            number_of_samples_bin = int(Decimal(bin_idx * df_results.shape[0]).to_integral_value(rounding=ROUND_HALF_UP))

            if number_of_samples_bin > 0:
                df_results_treated = df_results.iloc[0:number_of_samples_bin].loc[df_results[treatment_col] == 1].copy()
                df_results_control = df_results.iloc[0:number_of_samples_bin].loc[df_results[treatment_col] == 0].copy()

                r_ts.append(df_results_treated.loc[(df_results_treated[response_col] == 1)].shape[0])
                n_ts.append(df_results_treated.shape[0])
                r_cs.append(df_results_control.loc[(df_results_control[response_col] == 1)].shape[0])
                n_cs.append(df_results_control.shape[0])
            else:
                r_ts.append(0)
                n_ts.append(0)
                r_cs.append(0)
                n_cs.append(0)

        return UpliftEvaluation.calculate_uplift(r_ts, n_ts, r_cs, n_cs, bin_list, df_results.loc[df_results[treatment_col] == 1].shape[0])



import logging
import pickle
from datetime import datetime

import numpy as np
from causalml.inference.meta import BaseRClassifier
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from autoum.approaches.utils import ApproachParameters, DataSetsHelper, Helper


# TODO: Cost sensitive learning
# TODO: Feature Importance


class RLearner:
    """
    R-Learner proposed by Nie and Wager (2020)

    R-Learner is a two-step algorithm for heterogeneous treatment effect estimation.

    1. Step:
    Estimate main effect m*(x) (marginal effect) and treatment propensity e*(x) using (cross-validation) and any machine learning algorithm to form the "R-loss"

    * m*(x): (Informally, in my own words): Use all variables, except for the dependent treatment variable T, to predict Y
    * e*(x): (Informally, in my own words): Use all variables, except for the dependent response variable Y, to predict T (find confounders)

    2. Step:
    Optimize/Minimize the R-Loss Function using m* and e* using any machine learning algorithm
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for the R-Learner approach
        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """
        self.parameters = parameters
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using the R-Learner.

        Side note: For randomized controlled experiments, we can use the "observed propensity score", that is, the ratio between the individuals who were subject to a treatment
        and the individuals who were not subject to a treatment. For observational data, we need to estimate the propensity score.

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        # Add causalML's 'treatment' column
        experiment_groups_col = Helper.add_treatment_group_key(data_set_helper.df_train)

        # Initiate outcome and effect learner
        outcome_learner = RandomForestClassifier(**self.parameters)  # estimator for main effect and treatment propensity - stage 1. Should be a classifier
        effect_learner = RandomForestRegressor(**self.parameters)  # estimate the treatment effect - stage 2. Should be a regressor

        self.log.debug("Start fitting R-Learner ...")

        # Get the covariates for the data set
        x_train = data_set_helper.x_train
        x_test = data_set_helper.x_test

        # R Learner
        num_treatment_samples = np.count_nonzero(data_set_helper.df_train['treatment'].to_numpy() == 1)
        r_learner = BaseRClassifier(outcome_learner=outcome_learner, effect_learner=effect_learner, control_name='c', random_state=self.parameters["random_state"])
        r_learner.fit(X=x_train, treatment=experiment_groups_col, y=data_set_helper.y_train, p=Helper.get_propensity_score(x_train.shape[0], num_treatment_samples))

        self.log.debug(r_learner)

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_R-Learner_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(r_learner, open(filename, 'wb'))

        self.log.debug("Predicting ... ")

        score_train = r_learner.predict(X=x_train)
        score_test = r_learner.predict(X=x_test)

        if data_set_helper.valid:
            x_valid = data_set_helper.x_valid
            score_valid = r_learner.predict(X=x_valid)
        else:
            score_valid = np.array([0], np.int32)

        return {
            "score_train": score_train.flatten(),
            "score_valid": score_valid.flatten(),
            "score_test": score_test.flatten()
        }



import logging
import pickle
from datetime import datetime

import numpy as np
from sklearn.ensemble import RandomForestClassifier

from autoum.approaches.utils import ApproachParameters, DataSetsHelper, Helper


class Traditional:
    """
    Traditional Response Modeling Approach

    Here, only the treatment group is used for training the classifier
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for the traditional approach

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """
        self.parameters = parameters
        self.cost_sensitive = approach_parameters.cost_sensitive
        self.feature_importance = approach_parameters.feature_importance
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using the traditional approach.

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        # Training (Build dataset which contains only the treated individuals)
        df_train_treat = data_set_helper.df_train.loc[data_set_helper.df_train.treatment == 1]  # DataFrame
        x_train_treat = df_train_treat.drop(['response', 'treatment'], axis=1).to_numpy()  # Features
        y_train_treat = df_train_treat['response'].to_numpy()  # Target

        # Validation (Features)
        if data_set_helper.valid:
            x_valid = data_set_helper.df_valid.drop(['response', 'treatment'], axis=1).to_numpy()
        else:
            x_valid = None

        # Testing (Features)
        x_test = data_set_helper.df_test.drop(['response', 'treatment'], axis=1).to_numpy()
        # Training (for prediciton purposes we need the whole data set)
        x_train = data_set_helper.df_train.drop(['response', 'treatment'], axis=1).to_numpy()

        unique_classes = np.unique(y_train_treat)
        if len(unique_classes) < 2:
            self.log.error(f"Training set only contains samples from one class. Available Classes: {unique_classes}")
            raise Exception("Training set only contains samples from one class. Available Classes: {}".format(unique_classes))

        if self.cost_sensitive:
            # Calculate class weights
            class_weights = Helper.create_class_weight(y_train_treat)
            clf = RandomForestClassifier(class_weight=class_weights)
        else:
            clf = RandomForestClassifier()

        clf.set_params(**self.parameters)

        self.log.debug("Start fitting Traditional ...")

        clf.fit(x_train_treat, y_train_treat)

        self.log.debug(clf)

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_Tradtional_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(clf, open(filename, 'wb'))

        traditional_dict = {}

        if self.feature_importance:
            traditional_dict["feature_importance"] = clf.feature_importances_

        self.log.debug("Predicting ...")

        traditional_dict["score_train"] = clf.predict_proba(x_train)[:, 1]
        traditional_dict["score_test"] = clf.predict_proba(x_test)[:, 1]
        if x_valid is not None:
            traditional_dict["score_valid"] = clf.predict_proba(x_valid)[:, 1]
        else:
            traditional_dict["score_valid"] = []

        return traditional_dict



from os import path

import pandas as pd
from sklearn import preprocessing

pd.set_option('display.max_columns', 100)


class Hillstrom:

    def __init__(self, path_folder: str):

        self.hillstrom_original_path = path_folder + "hillstrom.csv"
        self.hillstrom_visit_path = path_folder + "Hillstrom_Email_visit.csv"
        self.hillstrom_w_visit_path = path_folder + "Hillstrom_Email_w_visit.csv"
        self.hillstrom_m_visit_path = path_folder + "Hillstrom_Email_m_visit.csv"
        self.hillstrom_conversion_path = path_folder + "Hillstrom_Email_conversion.csv"
        self.hillstrom_w_conversion_path = path_folder + "Hillstrom_Email_w_conversion.csv"
        self.hillstrom_m_conversion_path = path_folder + "Hillstrom_Email_m_conversion.csv"

    def prep(self, only_women: bool=False, only_men: bool=False, visit: bool=True):
        """
        Prepare the Hillstrom dataset and store the csv files in the filesystem.

        1. Resample both groups such that they have the same amount of items (# treated = # controlled)
        2. Apply LabelEncoder on categorical columns
        3. Create target column with a binary value

        :param only_women: True, if only th ewomen treatment should be considered. False otherwise.
        :param only_men: True, if only the men treatment should be considered. False otherwise.
        :param visit: True (default) if the feature "visit" shall be used as response variable.
            False if the feature "conversion" shall be the response variable. While visit indicates just a small response,
            conversion indicates that the customer actually bought products.

        """
        if not only_women and not only_men and not visit:
            if path.exists(self.hillstrom_conversion_path):
                return pd.read_csv(self.hillstrom_conversion_path)
        elif not only_women and not only_men and visit:
            if path.exists(self.hillstrom_visit_path):
                return pd.read_csv(self.hillstrom_visit_path)
        elif only_women and not only_men and visit:
            if path.exists(self.hillstrom_w_visit_path):
                return pd.read_csv(self.hillstrom_w_visit_path)
        elif only_women and not only_men and not visit:
            if path.exists(self.hillstrom_w_conversion_path):
                return pd.read_csv(self.hillstrom_w_conversion_path)
        elif not only_women and only_men and visit:
            if path.exists(self.hillstrom_m_visit_path):
                return pd.read_csv(self.hillstrom_m_visit_path)
        elif not only_women and only_men and not visit:
            if path.exists(self.hillstrom_m_conversion_path):
                return pd.read_csv(self.hillstrom_m_conversion_path)

        data = pd.read_csv(self.hillstrom_original_path)

        # 1. Create treatment column
        if only_women:
            data = data.loc[(data.segment != "Mens E-Mail")].copy()
            data["treatment"] = [0 if x == "No E-Mail" else 1 for x in data.segment]
        elif only_men:
            data = data.loc[(data.segment != "Womens E-Mail")].copy()
            data["treatment"] = [0 if x == "No E-Mail" else 1 for x in data.segment]
        else:
            data["treatment"] = [0 if x == "No E-Mail" else 1 for x in data.segment]

        data.drop((["segment"]), axis=1, inplace=True)

        ## 2. Apply LabelEncoder on categorical columns
        le = preprocessing.LabelEncoder()
        oe = preprocessing.OrdinalEncoder()
        data.history_segment = oe.fit_transform(data.history_segment.to_numpy().reshape(-1, 1)).astype('int32')
        data.zip_code = le.fit_transform(data.zip_code)
        data.channel = le.fit_transform(data.channel)

        ## 3. Create target column with a binary value
        if visit:
            data["response"] = [1 if x > 0.0 else 0 for x in data.visit]
        else:
            data["response"] = [1 if x > 0.0 else 0 for x in data.conversion]

        data.drop((["spend"]), axis=1, inplace=True)
        data.drop((["conversion"]), axis=1, inplace=True)
        data.drop((["visit"]), axis=1, inplace=True)

        # 4. Remove duplicates
        data.drop_duplicates(inplace=True, ignore_index=True)
        data.reset_index(inplace=True, drop=True)

        if only_women:
            if visit:
                save_path = self.hillstrom_w_visit_path
            else:
                save_path = self.hillstrom_w_conversion_path

        elif only_men:
            if visit:
                save_path = self.hillstrom_m_visit_path
            else:
                save_path = self.hillstrom_m_conversion_path

        else:
            if visit:
                save_path = self.hillstrom_visit_path
            else:
                save_path = self.hillstrom_conversion_path

        data.to_csv(save_path, index=False)

        return data



from os import path

import numpy as np
import pandas as pd
from sklearn import preprocessing

pd.set_option('display.max_columns', 100)
pd.set_option('display.width', 320)


class Bank_Telemarketing:

    def __init__(self, path_folder: str):
        ## Define paths
        self.bank_additional_original_path = path_folder + "bank-additional-full.csv"
        self.bank_additional_path_1 = path_folder + "Bank-Telemarketing_1.csv"
        self.bank_additional_path_2 = path_folder + "Bank-Telemarketing_2.csv"

    def prep_contact(self):
        """
        Prepare the bank_telemarketing dataset and store the csv files in the filesystem. Here only the contact column
        is used as the basis for creating the treatment column.

        Note: The duration column highly affects the output target (e.g., if duration=0 then y='no').
        Yet, the duration is not known before a call is performed. Also, after the end of the call y is obviously known.
        Thus, this input should only be included for benchmark purposes and should be discarded if the intention is to
        have a realistic predictive model.

        1. Apply LabelEncoder on categorical columns
        2. Encode treatment and resample both groups such that they have the same amount of items (# treated = # controlled)
        3. Create target column with a binary value
        4. Drop duplicates
        """

        if path.exists(self.bank_additional_path_1):
            return pd.read_csv(self.bank_additional_path_1)

        data = pd.read_csv(self.bank_additional_original_path, sep=';')

        # 1. Apply LabelEncoder on categorical columns
        # Find all columns to encode. All colunms of data type object will be encoded. All columns that should not be
        # encoded, must have a dtype different than object. Thus before using this method, please be sure to convert all
        # columns to the correct data type.
        encode_columns = []

        for col in data:
            if data[col].dtype == np.object:
                encode_columns.append(col)

        # Exclude treatment and control variable
        encode_columns.remove("contact")
        encode_columns.remove("y")

        le = preprocessing.LabelEncoder()
        for column in encode_columns:
            data[column] = le.fit_transform(data[column])

        # 2. Encode treatment
        # Cellular vs Telephone
        data["treatment"] = [0 if x == "telephone" else 1 for x in data.contact]
        data.drop((["contact"]), axis=1, inplace=True)

        # 3. Create target column with a binary value
        data["response"] = [1 if x == "yes" else 0 for x in data.y]
        data.drop((["y"]), axis=1, inplace=True)
        data.drop((["duration"]), axis=1, inplace=True)

        # 4. Remove duplicates
        data.drop_duplicates(inplace=True, ignore_index=True)
        data.reset_index(inplace=True, drop=True)

        data.to_csv(self.bank_additional_path_1, index=False)

        return data

    def prep_outcome(self):
        """
        Prepare the bank_telemarketing dataset and store the csv files in the filesystem. Here the contact column and
        the outcome column are used as the basis for creating the treatment column.
        Treatment = contacted cellular & not contacted in last campaign

        Note: The duration column highly affects the output target (e.g., if duration=0 then y='no').
        Yet, the duration is not known before a call is performed. Also, after the end of the call y is obviously known.
        Thus, this input should only be included for benchmark purposes and should be discarded if the intention is to
        have a realistic predictive model.

        1. Apply LabelEncoder on categorical columns
        2. Encode treatment and resample both groups such that they have the same amount of items (# treated = # controlled)
        3. Create target column with a binary value
        4. Drop duplicates
        """

        if path.exists(self.bank_additional_path_2):
            return pd.read_csv(self.bank_additional_path_2)

        data = pd.read_csv(self.bank_additional_original_path, sep=';')

        # 1. Apply LabelEncoder on categorical columns
        # Find all columns to encode. All colunms of data type object will be encoded. All columns that should not be
        # encoded, must have a dtype different than object. Thus before using this method, please be sure to convert all
        # columns to the correct data type.
        encode_columns = []

        for col in data:
            if data[col].dtype == np.object:
                encode_columns.append(col)

        # Exclude treatment and control variable
        encode_columns.remove("contact")
        encode_columns.remove("y")

        le = preprocessing.LabelEncoder()
        for column in encode_columns:
            data[column] = le.fit_transform(data[column])

        # 2. Encode treatment
        # Treated = Contacted via cellular and not contacted in last campaign
        data["contacted"] = [0 if x == "telephone" else 1 for x in data.contact]
        data["contacted_last_campaign"] = [1 if x == 1 else 0 for x in data.poutcome]
        data["treatment"] = data["contacted"] * data["contacted_last_campaign"]

        data.drop((["contact"]), axis=1, inplace=True)
        data.drop((["contacted"]), axis=1, inplace=True)
        data.drop((["contacted_last_campaign"]), axis=1, inplace=True)

        # 3. Create target column with a binary value
        data["response"] = [1 if x == "yes" else 0 for x in data.y]
        data.drop((["y"]), axis=1, inplace=True)
        data.drop((["duration"]), axis=1, inplace=True)

        # 4. Remove duplicates
        data.drop_duplicates(inplace=True, ignore_index=True)
        data.reset_index(inplace=True, drop=True)

        data.to_csv(self.bank_additional_path_2, index=False)

        return data



import logging
import pickle
from datetime import datetime

from sklearn.ensemble import RandomForestClassifier

from autoum.approaches.utils import ApproachParameters, DataSetsHelper, Helper


class SLearner:
    """
    S-Learner proposed in multiple papers (e.g., Künzel et al. 2019)

    Here, the treatment variable is used as another dependent variable during training.
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for the S-Learner

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """
        self.parameters = parameters
        self.cost_sensitive = approach_parameters.cost_sensitive
        self.feature_importance = approach_parameters.feature_importance
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using the S-Learner.

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        if self.cost_sensitive:
            # Calculate class weights
            class_weights = Helper.create_class_weight(data_set_helper.y_train)
            clf = RandomForestClassifier(class_weight=class_weights)
        else:
            clf = RandomForestClassifier()

        clf.set_params(**self.parameters)

        self.log.debug("Start fitting S-Learner ...")

        # Create data frame including the treatment variable but not response variable
        df_train_with_treatment = data_set_helper.df_train.drop(['response'], axis=1)
        df_test_with_treatment = data_set_helper.df_test.drop(['response'], axis=1)

        clf.fit(df_train_with_treatment.to_numpy(), data_set_helper.y_train)

        self.log.debug(clf)

        response_dict = {}

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_S-Learner_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(clf, open(filename, 'wb'))

        if self.feature_importance:
            response_dict["feature_importance"] = clf.feature_importances_

        self.log.debug("Predicting ...")

        score_dict = {}
        df_train_with_treatment["treatment"] = 0
        score_dict["score_train_t_0"] = clf.predict_proba(df_train_with_treatment.to_numpy())[:, 1]
        df_train_with_treatment["treatment"] = 1
        score_dict["score_train_t_1"] = clf.predict_proba(df_train_with_treatment.to_numpy())[:, 1]
        response_dict["score_train"] = score_dict["score_train_t_1"] - score_dict["score_train_t_0"]

        df_test_with_treatment["treatment"] = 0
        score_dict["score_test_t_0"] = clf.predict_proba(df_test_with_treatment.to_numpy())[:, 1]
        df_test_with_treatment["treatment"] = 1
        score_dict["score_test_t_1"] = clf.predict_proba(df_test_with_treatment.to_numpy())[:, 1]
        response_dict["score_test"] = score_dict["score_test_t_1"] - score_dict["score_test_t_0"]

        if data_set_helper.valid:
            df_valid_with_treatment = data_set_helper.df_valid.drop(['response'], axis=1)
            df_valid_with_treatment["treatment"] = 0
            score_dict["score_valid_t_0"] = clf.predict_proba(df_valid_with_treatment.to_numpy())[:, 1]
            df_valid_with_treatment["treatment"] = 1
            score_dict["score_valid_t_1"] = clf.predict_proba(df_valid_with_treatment.to_numpy())[:, 1]
            response_dict["score_valid"] = score_dict["score_valid_t_1"] - score_dict["score_valid_t_0"]

        else:
            response_dict["score_valid"] = []

        return response_dict



import numpy as np
import pandas as pd
from sklearn.utils import class_weight


class Helper:
    """
    Helper class for approaches
    """

    @staticmethod
    def create_class_weight(y):
        """
        Create class weights for each class

        :param y: Target column
        :return: A dictionary containing the weights for each class
        """
        unique_classes = np.unique(y)
        class_weights = class_weight.compute_class_weight('balanced', classes=unique_classes, y=y)
        d_class_weights = {}
        for i in range(len(unique_classes)):
            d_class_weights[unique_classes[i]] = class_weights[i]

        return d_class_weights

    @staticmethod
    def add_treatment_group_key(df: pd.DataFrame) -> np.array:
        """
        CausalML expects a "treatment" column which contains a unique string value for individuals belonging to the control group and a unique string value for individuals
        belonging to the treatment group.

        :param df: Dataframe that shall be reformatted
        :return: Array containing String values for the control and the treatment group
        """

        experiment_groups_col = ["c" if x == 0 else "t" for x in df.treatment]
        experiment_groups_col = np.array(experiment_groups_col)

        return experiment_groups_col

    @staticmethod
    def get_propensity_score(length: int, num_treatment_samples: int = 0):
        """
        Return propensity score.
        The propensity score is defined as the probability of receiving a treatment (at least in randomized, controlled experiments)

        :param length: Length of the array
        :param num_treatment_samples: Number of treatment samples in the training set. If None, return 0.5 as propensity score (test set)
        :return: Propensity Score for each
        """

        if num_treatment_samples > 0:
            propensity_score = np.empty(length)
            propensity_score.fill(num_treatment_samples / length)
        else:
            propensity_score = np.empty(length)
            propensity_score.fill(0.5)

        return propensity_score


class DataSetsHelper:
    """
    Wrapper class that saves a training, test and validation set for given training, test and validation DataFrames.
    """

    def __init__(self, df_train: pd.DataFrame, df_valid: pd.DataFrame, df_test: pd.DataFrame):
        """
        Creates an object containing the training, validation, and test set for a given dataset. This class contains utility methods for getting the full test, training and
        validation sets (including the treatment and response indicator), covariates and target sets.

        :param df_train: Dataframe containing the training set (including the treatment and response indicator)
        :param df_train: Dataframe containing the validation set (including the treatment and response indicator)
        :param df_test: Dataframe containing the test set (including the treatment and response indicator)
        """
        # Training
        self.df_train = df_train                                                        # Full DataFrame
        self.x_train = df_train.drop(['response', 'treatment'], axis=1).to_numpy()      # Covariates
        self.y_train = df_train['response'].to_numpy()                                  # Target

        # Test
        self.df_test = df_test                                                          # Full DataFrame
        self.x_test = df_test.drop(['response', 'treatment'], axis=1).to_numpy()        # Covariates
        self.y_test = df_test['response'].to_numpy()                                    # Target

        # Validation
        self.valid = True
        if df_valid.empty:
            self.valid = False
        else:
            self.df_valid = df_valid                                                    # Full DataFrame
            self.x_valid = df_valid.drop(['response', 'treatment'], axis=1).to_numpy()  # Covariates
            self.y_valid = df_valid['response'].to_numpy()                              # Target

    def get_covariates(self):
        return self.x_train, self.x_valid, self.x_test

    def get_targets(self):
        return self.y_train, self.y_valid, self.y_test

    def get_dataframes(self):
        return self.df_train, self.df_valid, self.df_test


class ApproachParameters:
    """
    Utility class that encompassees all parameters needed to create an approach instance.
    """

    def __init__(self, cost_sensitive: bool, feature_importance: bool, path: str, save: bool, split_number: int):
        """
        Utility class that encompassees all parameters needed to create an approach instance.

        :param cost_sensitive: Set this to true for cost sensitive learning.
        :param feature_importance: Set this to True to return the feature importances of the classifiers
        :param path: Path where the models generated during approach execution should be saved.
        :param save: Set this to True if during training the generated models should be saved
        :param split_number: Number of current split. Used for logging and saving purposes in order to identify saved models.

        """
        self.cost_sensitive = cost_sensitive
        self.feature_importance = feature_importance
        self.path = path
        self.save = save
        self.split_number = split_number



import logging
import os
import time
from collections import defaultdict
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from autoum.approaches.bayesian_causal_forest import BayesianCausalForest
from autoum.approaches.class_variable_transformation import ClassVariableTransformation
from autoum.approaches.generalized_random_forest import GeneralizedRandomForest
from autoum.approaches.lais_generalization import LaisGeneralization
from autoum.approaches.r_learner import RLearner
from autoum.approaches.s_learner import SLearner
from autoum.approaches.traditional import Traditional
from autoum.approaches.treatment_dummy import TreatmentDummy
from autoum.approaches.two_model import TwoModel
from autoum.approaches.uplift_random_forest import UpliftRandomForest
from autoum.approaches.utils import ApproachParameters, DataSetsHelper
from autoum.approaches.x_learner import XLearner
from autoum.const.const import *
from autoum.datasets.utils import get_data_home

logging = logging.getLogger(__name__)


class HelperPipeline:
    """
    Helper class with utility functions which support the Pipelines.
    """

    def __init__(self):
        pass

    @staticmethod
    def apply_approach(classifier, data_set_helper: DataSetsHelper, feature_importance: bool):
        """
        You can pass classifiers for every approach:
         - S-Learner
         - Traditional
         - Uplift Random Forest (with 8 different splitting criteria)
         - Class Variable Transformation
         - Lais Generalization
         - Two-Model
         - X-Learner
         - R-Learner
         - Treatment Dummy
         - Bayesian Causal Forest
         - Generalized Random Forest

        For parameter definition see :func:`~HelperPipeline.apply_uplift_approaches`
        """
        feature_importances_dict = {}
        logging.info(str(classifier.__class__.__name__))

        try:
            # Predict training, validation and test scores
            result_dict = classifier.analyze(data_set_helper)

            score_uplift_train = result_dict["score_train"]
            score_uplift_valid = result_dict["score_valid"]
            score_uplift_test = result_dict["score_test"]

            if feature_importance:
                feature_importances_dict = result_dict["feature_importance"]

        except Exception as e:
            logging.error(f"Error in {str(classifier.__class__.__name__)} Approach; the approach is no longer considered for further evaluation; error message: \n {e}")
            score_uplift_train, score_uplift_valid, score_uplift_test = np.nan, np.nan, np.nan

        return score_uplift_train, score_uplift_valid, score_uplift_test, feature_importances_dict

    @staticmethod
    def apply_uplift_approaches(df_train: pd.DataFrame,
                                df_valid: pd.DataFrame,
                                df_test: pd.DataFrame,
                                parameters: dict,
                                approach: list,
                                split_number: int,
                                cost_sensitive: bool = False,
                                feature_importance: bool = False,
                                save_models: bool = False) -> dict:
        """
        Apply given uplift modeling approaches on the given dataframes and return the scores

        :param df_train: Dataframe containing the training set including all covariates, response, and treatment variable.
        :param df_valid: Dataframe containing the validation set including all covariates, response, and treatment variable. Note, this DataFrame can also be empty for final
            model creation.
        :param df_test: Dataframe containing the test set including all covariates, response, and treatment variable.
        :param parameters: Parameter dictionary which contains the dictionaries of the various uplift modeling approaches
        :param approach: List containing the names of the approaches which should be evaluated.
        :param split_number: Number of current split. Used for logging and saving purposes in order to identify saved models.
        :param cost_sensitive: Set this to true for cost sensitive learning.
        :param feature_importance: Set this to True to return the feature importances of the classifiers
        :param save_models: True if the models generated during training shall be saved. False otherwise.
        :return: Dictionary with the following keys: df_scores_train, df_scores_test, df_train, df_test, feature_importances (empty dictionary if not used)
        """

        # Create DataFrames which contain results (i.e., the uplift scores)
        df_scores_train = pd.DataFrame()
        df_scores_valid = pd.DataFrame()
        df_scores_test = pd.DataFrame()

        # Dictionary which will be used for storing the feature importance of each classifier
        fid = {}

        # DataSetsHelper contains all base dataframes and their transformations. These are used in the analyze methods of the approaches.
        ds_helper = DataSetsHelper(df_train=df_train, df_valid=df_valid, df_test=df_test)
        # ApproachParameters contains alll parameters necessary to initialize an approach classifier
        root = f"{get_data_home()}/models/"
        approach_params = ApproachParameters(cost_sensitive=cost_sensitive, feature_importance=feature_importance, path=root, save=save_models, split_number=split_number)

        # This dictionary is used as wrapper for passing all parameters at once for apply_approach
        apply_params = {
            "data_set_helper": ds_helper,
            "feature_importance": feature_importance,
        }

        ### Apply Uplift Modeling Approaches ###
        start = time.time()

        # 1. UPLIFT RANDOM FOREST APPROACH
        # 1.1 WITH EUCLIDEAN DISTANCE
        if "URF_ED" in approach:
            urf_ed = UpliftRandomForest(parameters[URF_TITLE + "_parameters"], approach_params, eval_function="ED")
            df_scores_train[URF_ED_TITLE], df_scores_valid[URF_ED_TITLE], df_scores_test[URF_ED_TITLE], _ = HelperPipeline.apply_approach(urf_ed, **apply_params)

        # 1.2 WITH KULLBACK-LEIBLER DIVERGENCE
        if "URF_KL" in approach:
            urf_kl = UpliftRandomForest(parameters[URF_TITLE + "_parameters"], approach_params, eval_function="KL")
            df_scores_train[URF_KL_TITLE], df_scores_valid[URF_KL_TITLE], df_scores_test[URF_KL_TITLE], _ = HelperPipeline.apply_approach(urf_kl, **apply_params)

        # 1.3 WITH CHI-SQAURED DIVERGENCE
        if "URF_CHI" in approach:
            urf_chi = UpliftRandomForest(parameters[URF_TITLE + "_parameters"], approach_params, eval_function="Chi")
            df_scores_train[URF_CHI_TITLE], df_scores_valid[URF_CHI_TITLE], df_scores_test[URF_CHI_TITLE], _ = HelperPipeline.apply_approach(urf_chi, **apply_params)

        # 1.4 WITH DELTA-DELTA-PI Criterion
        if "URF_DDP" in approach:
            urf_ddp = UpliftRandomForest(parameters[URF_TITLE + "_parameters"], approach_params, eval_function="DDP")
            df_scores_train[URF_DDP_TITLE], df_scores_valid[URF_DDP_TITLE], df_scores_test[URF_DDP_TITLE], _ = HelperPipeline.apply_approach(urf_ddp, **apply_params)

        # 1.5 WITH CONTEXTUAL TREATMENT SELECTION
        if "URF_CTS" in approach:
            urf_cts = UpliftRandomForest(parameters[URF_TITLE + "_parameters"], approach_params, eval_function="CTS")
            df_scores_train[URF_CTS_TITLE], df_scores_valid[URF_CTS_TITLE], df_scores_test[URF_CTS_TITLE], _ = HelperPipeline.apply_approach(urf_cts, **apply_params)

        # 1.6 IT CRITERION
        if "URF_IT" in approach:
            urf_it = UpliftRandomForest(parameters[URF_TITLE + "_parameters"], approach_params, eval_function="IT")
            df_scores_train[URF_IT_TITLE], df_scores_valid[URF_IT_TITLE], df_scores_test[URF_IT_TITLE], _ = HelperPipeline.apply_approach(urf_it, **apply_params)

        # 1.7 CIT CRITERION
        if "URF_CIT" in approach:
            urf_cit = UpliftRandomForest(parameters[URF_TITLE + "_parameters"], approach_params, eval_function="CIT")
            df_scores_train[URF_CIT_TITLE], df_scores_valid[URF_CIT_TITLE], df_scores_test[URF_CIT_TITLE], _ = HelperPipeline.apply_approach(urf_cit, **apply_params)

        # 2. S-LEARNER
        if "SLEARNER" in approach:
            s_learner = SLearner(parameters[SLEARNER_TITLE + "_parameters"], approach_params)
            df_scores_train[SLEARNER_TITLE], df_scores_valid[SLEARNER_TITLE], df_scores_test[SLEARNER_TITLE], fid[SLEARNER_TITLE] = HelperPipeline.apply_approach(s_learner,
                                                                                                                                                                  **apply_params)

        # 3. CLASS VARIABLE TRANSFORMATION
        if "CVT" in approach:
            cvt = ClassVariableTransformation(parameters[CVT_TITLE + "_parameters"], approach_params)
            df_scores_train[CVT_TITLE], df_scores_valid[CVT_TITLE], df_scores_test[CVT_TITLE], fid[CVT_TITLE] = HelperPipeline.apply_approach(cvt, **apply_params)

        # 4. LAIS OUTCOME APPROACH
        if "LAIS" in approach:
            lais = LaisGeneralization(parameters[LAIS_TITLE + "_parameters"], approach_params)
            df_scores_train[LAIS_TITLE], df_scores_valid[LAIS_TITLE], df_scores_test[LAIS_TITLE], fid[LAIS_TITLE] = HelperPipeline.apply_approach(lais, **apply_params)

        # 5. TWO-MODEL APPROACH
        if "TWO_MODEL" in approach:
            two_model = TwoModel(parameters[TWO_MODEL_TITLE + "_parameters"], approach_params)
            df_scores_train[TWO_MODEL_TITLE], df_scores_valid[TWO_MODEL_TITLE], df_scores_test[TWO_MODEL_TITLE], fid_tm = HelperPipeline.apply_approach(two_model, **apply_params)

            fid.update(fid_tm)

        # 6. TRADITIONAL APPROACH
        if "TRADITIONAL" in approach:
            traditional = Traditional(parameters[TRADITIONAL_TITLE + "_parameters"], approach_params)
            df_scores_train[TRADITIONAL_TITLE], df_scores_valid[TRADITIONAL_TITLE], df_scores_test[TRADITIONAL_TITLE], fid[TRADITIONAL_TITLE] = HelperPipeline.apply_approach(
                traditional, **apply_params)

        # 7. X-LEARNER
        if "XLEARNER" in approach:
            x_learner = XLearner(parameters[XLEARNER_TITLE + "_parameters"], approach_params)
            df_scores_train[XLEARNER_TITLE], df_scores_valid[XLEARNER_TITLE], df_scores_test[XLEARNER_TITLE], _ = HelperPipeline.apply_approach(x_learner, **apply_params)

        # 8. R-LEARNER
        if "RLEARNER" in approach:
            r_learner = RLearner(parameters[RLEARNER_TITLE + "_parameters"], approach_params)
            df_scores_train[RLEARNER_TITLE], df_scores_valid[RLEARNER_TITLE], df_scores_test[RLEARNER_TITLE], _ = HelperPipeline.apply_approach(r_learner, **apply_params)

        # 9. TREATMENT DUMMY
        if "TREATMENT_DUMMY" in approach:
            tda = TreatmentDummy(parameters[TREATMENT_DUMMY_TITLE + "_parameters"], approach_params)
            df_scores_train[TREATMENT_DUMMY_TITLE], df_scores_valid[TREATMENT_DUMMY_TITLE], df_scores_test[TREATMENT_DUMMY_TITLE], _ = HelperPipeline.apply_approach(tda,
                                                                                                                                                                     **apply_params)

        # 10. GENERALIZED RANDOM FOREST
        if "GRF" in approach:
            grf = GeneralizedRandomForest(parameters[GRF_TITLE + "_parameters"], approach_params)
            df_scores_train[GRF_TITLE], df_scores_valid[GRF_TITLE], df_scores_test[GRF_TITLE], _ = HelperPipeline.apply_approach(grf, **apply_params)

        # 11. BAYESIAN CAUSAL FOREST
        if "BCF" in approach:
            bcf = BayesianCausalForest(parameters[BCF_TITLE + "_parameters"], approach_params)
            df_scores_train[BCF_TITLE], df_scores_valid[BCF_TITLE], df_scores_test[BCF_TITLE], _ = HelperPipeline.apply_approach(bcf, **apply_params)

        appr_name = approach[0]

        end = time.time()
        logging.info(f'Function {appr_name} took {(end - start):.2f} s')

        df_scores_train['response'] = df_train['response'].to_numpy()
        df_scores_train['treatment'] = df_train['treatment'].to_numpy()

        df_scores_valid['response'] = df_valid['response'].to_numpy()
        df_scores_valid['treatment'] = df_valid['treatment'].to_numpy()

        df_scores_test['response'] = df_test['response'].to_numpy()
        df_scores_test['treatment'] = df_test['treatment'].to_numpy()

        return {
            "df_scores_train": df_scores_train,
            "df_scores_valid": df_scores_valid,
            "df_scores_test": df_scores_test,
            "feature_importances": fid
        }

    @staticmethod
    def cast_to_dataframe(_ds: list):
        """
        Cast the given dictionaries to one dataframe

        :param _ds: List of dictionaries
        :return: DataFrame
        """

        if isinstance(_ds, list):
            final_dict = defaultdict(list)

            # Iterate through each dictionary in ds
            for d in _ds:
                for key, value in d.items():
                    final_dict[key].append(value)
            # Cast back to dict
            final_dict = dict(final_dict)

            df = pd.DataFrame.from_dict(final_dict)
        else:
            df = pd.DataFrame(_ds, index=[0])
        return df

    @staticmethod
    def save_feature_importance(importances, feature_names, title, save_figure, plot_figure, path):
        """
        Save feature importance plot

        :param importances: Importances
        :param feature_names: Names of all features
        :param title: Title of the plot
        :param save_figure: True, if figure should be saved. False otherwise
        :param plot_figure: True, if figure should be plotted. False otherwise
        :param path: Path were the plot should be stores.
        """
        indices = np.argsort(importances)

        if plot_figure:
            plt.figure(figsize=(14, 8))
            plt.title(title)
            plt.barh(range(len(indices)), importances[indices], color='b', align='center')
            plt.yticks(range(len(indices)), [feature_names[i] for i in indices])
            plt.xlabel('Relative Importance')

        now = datetime.now()
        now_date = now.date()
        if save_figure:
            if not os.path.exists(path):
                os.makedirs(path)
            plt.savefig(path + title + "_" + str(now_date) + ".png")

        if plot_figure:
            plt.show()



import logging
import pickle
from datetime import datetime

import numpy as np
from econml.dml import CausalForestDML
from sklearn.linear_model import LassoCV

from autoum.approaches.utils import ApproachParameters, DataSetsHelper


# TODO: Cost sensitive learning
# TODO: Feature Importance


class GeneralizedRandomForest:
    """
    Generalized Random Forest proposed by Athey et al. (2019)
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for the generalized random forest (Athey et al., 2019)

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """

        self.parameters = parameters
        self.parameters["model_t"] = LassoCV()
        self.parameters["model_y"] = LassoCV()
        self.feature_importance = approach_parameters.feature_importance
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using generalized random forest

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        # Initiate CausalForestDML
        causal_forrest_classifier = CausalForestDML(**self.parameters)

        treated = data_set_helper.df_train["treatment"].copy()

        self.log.debug("Start fitting Generalized Random Forest ...")
        causal_forrest_classifier.fit(Y=data_set_helper.y_train, T=treated, X=data_set_helper.x_train, W=None)

        self.log.debug(causal_forrest_classifier)

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_GRF_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(causal_forrest_classifier, open(filename, 'wb'))

        self.log.debug("Predicting ... ")

        # Note: The .predict() method returns an ndarray in which each column contains the predicted uplift if the unit was in the corresponding treatment group.
        score_train = causal_forrest_classifier.effect(data_set_helper.x_train)
        score_test = causal_forrest_classifier.effect(data_set_helper.x_test)
        if data_set_helper.valid:
            score_valid = causal_forrest_classifier.effect(data_set_helper.x_valid)
        else:
            score_valid = np.array([0], np.int32)

        return {
            "score_train": score_train.flatten(),
            "score_valid": score_valid.flatten(),
            "score_test": score_test.flatten()
        }



"""
Base IO code for all datasets
"""

import gzip
import shutil
import tarfile
from os import environ, makedirs, path
from os.path import expanduser, join
from zipfile import ZipFile

from autoum.datasets.bank_telemarketing import Bank_Telemarketing
from autoum.datasets.criteo import Criteo
from autoum.datasets.criteo_v2 import CriteoV2
from autoum.datasets.hillstrom import Hillstrom
from autoum.datasets.lenta import Lenta
from autoum.datasets.socialpressure import SocialPressure
from autoum.datasets.starbucks import Starbucks


def get_data_home(data_home: str=None) -> str:
    """Return the path of the autoum data directory.

    This folder is used by some large dataset loaders to avoid downloading the data several times.
    By default, the data directory is set to a folder named 'autoum_data' in the user home folder.

    Alternatively, it can be set by the 'AUTOUM_DATA' environment variable or programmatically by giving an explicit folder path.
    The '~' symbol is expanded to the user home folder.

    If the folder does not already exist, it is automatically created.

    :param data_home: The path to autoum's data directory. If `None`, the default path is `~/sklearn_learn_data`.
    :return: The path to autoum data directory
    """
    if data_home is None:
        data_home = environ.get("AUTOUM_DATA", join("~", "autoum_output"))
    data_home = expanduser(data_home)
    makedirs(data_home, exist_ok=True)
    return data_home


def unpack_zip(zip_file_path: str, dest_path: str):
    """
    Unpack a zip file

    :param zip_file_path: Path of the zip file
    :param dest_path: Path where the zip file should be unzipped
    """
    if not path.isdir(dest_path):
        makedirs(dest_path, exist_ok=True)
        with ZipFile(zip_file_path, 'r') as zip_file:
            zip_file.extractall(dest_path)


def unpack_gz(gz_file_path: str, dest_path: str, dest_file: str):
    """
    Unpack a gz file

    :param gz_file_path: Path of the gz file
    :param dest_path: Path of the directory where the gz file should be unzipped
    :param dest_file: Name of the unpacked file
    """
    if not path.exists(dest_path + dest_file):
        makedirs(dest_path, exist_ok=True)
        with gzip.open(gz_file_path, 'rb') as f_in:
            with open(dest_path + dest_file, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)


def unpack_tar(tar_file_path: str, dest_path: str):
    """
    Unpack a tar file

    :param tar_file_path: Path of the tar file
    :param dest_path: Path where the tar file should be unzipped
    """
    if not path.isdir(dest_path):
        makedirs(dest_path, exist_ok=True)
        with tarfile.open(tar_file_path) as my_tar:
            my_tar.extractall(dest_path)


def get_hillstrom_women_visit():
    """Return Hillstrom dataset with women only and visit as response variable"""
    return get_hillstrom(only_women=True, only_men=False, visit=True)


def get_hillstrom_women_conversion():
    """Return Hillstrom dataset with women only and conversion as response variable"""
    return get_hillstrom(only_women=True, only_men=False, visit=False)


def get_hillstrom_men_visit():
    """Return Hillstrom dataset with men only and visit as response variable"""
    return get_hillstrom(only_women=False, only_men=True, visit=True)


def get_hillstrom_men_conversion():
    """Return Hillstrom dataset with men only and conversion as response variable"""
    return get_hillstrom(only_women=False, only_men=True, visit=False)


def get_hillstrom_visit():
    """Return Hillstrom dataset with men and women and visit as response variable"""
    return get_hillstrom(only_women=False, only_men=False, visit=True)


def get_hillstrom_conversion():
    """Return Hillstrom dataset with men and women and conversion as response variable"""
    return get_hillstrom(only_women=False, only_men=False, visit=False)


def get_hillstrom(only_women: bool, only_men: bool, visit: bool):
    """
    Get the Hillstrom dataset

    :param only_women: True, if only the women treatment should be considered. False otherwise.
    :param only_men: True, if only the men treatment should be considered. False otherwise.
    :param visit: True, if the feature "visit" shall be used as response variable. False if the feature "conversion" shall be used as response variable.
    """
    # Unzip
    hillstrom_path_zip = "autoum/datasets/data/hillstrom.csv.zip"
    path_folder = get_data_home() + "/data/hillstrom-email/"
    unpack_zip(hillstrom_path_zip, path_folder)

    # Prepare
    hillstrom = Hillstrom(path_folder)
    data = hillstrom.prep(only_women=only_women, only_men=only_men, visit=visit)

    return data


def get_bank_telemarketing(prep_contact: bool):
    """
    Get the Bank Telemarketing dataset

    :param prep_contact: True, if the "contract" feature shall be used as repsonse. False, if the "contact" and "outcome" column shall be used as response
    """
    # Unzip
    bank_telemarketing_path_zip = "autoum/datasets/data/bank-additional-full.csv.zip"
    path_folder = get_data_home() + "/data/bank-telemarketing/"
    unpack_zip(bank_telemarketing_path_zip, path_folder)

    # Prepare
    bank_telemarketing = Bank_Telemarketing(path_folder)
    if prep_contact:
        data = bank_telemarketing.prep_contact()
    else:
        data = bank_telemarketing.prep_outcome()

    return data


def get_bank_telemarketing_contact():
    """Return Bank Telemarketing dataset with contact as response variable"""
    return get_bank_telemarketing(prep_contact=True)


def get_bank_telemarketing_outcome():
    """Return Bank Telemarketing dataset with contact and outcome as response variable"""
    return get_bank_telemarketing(prep_contact=False)


def get_criteo_v1(resample: bool):
    """
    Get the Criteo V1 dataset

    :param resample: True, if the dataset should be resampled. False otherwise
    """
    # Unzip
    criteo_path_zip = "autoum/datasets/data/criteo-uplift.csv.gz"
    path_folder = get_data_home() + "/data/criteo-v1/"
    file_name = "criteo-uplift.csv"
    unpack_gz(criteo_path_zip, path_folder, file_name)

    # Prepare
    criteo = Criteo(path_folder)
    if resample:
        return criteo.prep(resample=True)
    else:
        return criteo.prep(resample=False)


def get_criteo_v1_full():
    """Return Criteo V1 dataset"""
    return get_criteo_v1(resample=False)


def get_criteo_v1_resampled():
    """Return Criteo V1 dataset with resampling"""
    return get_criteo_v1(resample=True)


def get_criteo_v2(resample: bool):
    """
    Get the Criteo V2 dataset

    :param resample: True, if the dataset should be resampled. False otherwise
    """
    # Unzip
    criteo_path_zip = "autoum/datasets/data/criteo-uplift-v2.1.csv.gz"
    path_folder = get_data_home() + "/data/criteo-v2/"
    file_name = "criteo-uplift-v2.1.csv"
    unpack_gz(criteo_path_zip, path_folder, file_name)

    # Prepare
    criteo = CriteoV2(path_folder)
    if resample:
        return criteo.prep(resample=True)
    else:
        return criteo.prep(resample=False)


def get_criteo_v2_full():
    """Return Criteo V2 dataset"""
    return get_criteo_v2(resample=False)


def get_criteo_v2_resampled():
    """Return Criteo V2 dataset with resampling"""
    return get_criteo_v2(resample=True)


def get_lenta():
    """
    Get the Lenta dataset
    """
    # Unzip
    lenta_path_zip = "autoum/datasets/data/lenta_dataset.csv.gz"
    path_folder = get_data_home() + "/data/lenta/"
    file_name = "lenta_dataset.csv"
    unpack_gz(lenta_path_zip, path_folder, file_name)

    # Prepare
    lenta = Lenta(path_folder)
    return lenta.prep()


def get_social_pressure():
    """
    Get the Social Pressure dataset
    """
    # Unzip
    social_pressure_path_zip = "autoum/datasets/data/data.zip"
    path_folder = get_data_home() + "/data/social-pressure/"
    unpack_zip(social_pressure_path_zip, path_folder)

    # Prepare
    social_pressure = SocialPressure(path_folder)
    return social_pressure.prep()


def get_starbucks():
    """
    Get the Starbucks dataset
    """
    # Unzip
    starbucks_path_zip = "autoum/datasets/data/data.tar.gz"
    path_folder = get_data_home() + "/data/starbucks/"
    unpack_tar(starbucks_path_zip, path_folder)

    # Prepare
    starbucks = Starbucks(path_folder)
    return starbucks.prep()



"""
File containing constants
"""

############################## APPROACH NAMES ##############################

BCF_TITLE = "BCF"
CVT_TITLE = "CVT"
GRF_TITLE = "GRF"
LAIS_TITLE = "Lais"
RLEARNER_TITLE = "R_Learner"
SLEARNER_TITLE = "S_Learner"
TRADITIONAL_TITLE = "Traditional"
TREATMENT_DUMMY_TITLE = "Treatment_Dummy"
TWO_MODEL_TITLE = "Two_Model"
URF_TITLE = "URF"
URF_ED_TITLE = "URF_ED"
URF_KL_TITLE = "URF_KL"
URF_CHI_TITLE = "URF_CHI"
URF_DDP_TITLE = "URF_DDP"
URF_CTS_TITLE = "URF_CTS"
URF_IT_TITLE = "URF_IT"
URF_CIT_TITLE = "URF_CIT"
XLEARNER_TITLE = "X_Learner"

############################## COLOR SCHEMA ##############################
COLOR_SCHEMA = {
    BCF_TITLE: "#4daf4a",
    CVT_TITLE: "#984ea3",
    GRF_TITLE: "#00E307",
    LAIS_TITLE: "#6A0213",
    RLEARNER_TITLE: "#9400E6",
    SLEARNER_TITLE: "#003C86",
    TRADITIONAL_TITLE: "#e41a1c",
    TREATMENT_DUMMY_TITLE: "#FFDC3D",
    TWO_MODEL_TITLE: "#008607",
    URF_KL_TITLE: "#008169",
    URF_ED_TITLE: "#EF0096",
    URF_CHI_TITLE: "#00DCB5",
    URF_CTS_TITLE: "#FFCFE2",
    URF_DDP_TITLE: "#7CFFFA",
    URF_IT_TITLE: "#009FFA",
    URF_CIT_TITLE: "#68023F",
    XLEARNER_TITLE: "#F60239"
}

############################## GRAY COLOR SCHEMA ##############################
COLOR_SCHEMA_GRAY = {
    BCF_TITLE: "gray",
    CVT_TITLE: "silver",
    GRF_TITLE: "gainsboro",
    LAIS_TITLE: "whitesmoke",
    RLEARNER_TITLE: "darkgrey",
    SLEARNER_TITLE: "lightgray",
    TRADITIONAL_TITLE: "dimgray",
    TREATMENT_DUMMY_TITLE: "gray",
    TWO_MODEL_TITLE: "darkgray",
    URF_KL_TITLE: "#008169",
    URF_ED_TITLE: "#EF0096",
    URF_CHI_TITLE: "#00DCB5",
    URF_CTS_TITLE: "#FFCFE2",
    URF_DDP_TITLE: "#7CFFFA",
    URF_IT_TITLE: "#009FFA",
    URF_CIT_TITLE: "#68023F",
    XLEARNER_TITLE: "slategray"
}

############################## LINESTYLES ##############################
LINESTYLES = {
    BCF_TITLE: (0, (1, 1)),
    CVT_TITLE: (0, (1, 1)),
    GRF_TITLE: (0, (1, 1)),
    LAIS_TITLE: (0, (1, 1)),
    RLEARNER_TITLE: (0, (1, 1)),
    SLEARNER_TITLE: (0, (1, 10)),
    TRADITIONAL_TITLE: (0, (1, 1)),
    TREATMENT_DUMMY_TITLE: (0, (3, 1, 1, 1)),
    TWO_MODEL_TITLE: (0, (1, 1)),
    URF_KL_TITLE: (0, (1, 1)),
    URF_ED_TITLE: (0, (5, 10)),
    URF_CHI_TITLE: (0, (5, 1)),
    URF_CTS_TITLE: (0, (3, 10, 1, 10)),
    URF_DDP_TITLE: (0, (3, 5, 1, 5)),
    URF_IT_TITLE: (0, (1, 1)),
    URF_CIT_TITLE: (0, (1, 1)),
    XLEARNER_TITLE: (0, (1, 1))
}

############################## ABBREVIATIONS ##############################
NAMING_SCHEMA = {
    BCF_TITLE: "BCF",
    CVT_TITLE: "CVT",
    GRF_TITLE: "GRF",
    LAIS_TITLE: "LG",
    RLEARNER_TITLE: "RL",
    SLEARNER_TITLE: "SL",
    TRADITIONAL_TITLE: TRADITIONAL_TITLE,
    TREATMENT_DUMMY_TITLE: "TDA",
    TWO_MODEL_TITLE: "TM",
    URF_KL_TITLE: "U-KL",
    URF_ED_TITLE: "U-ED",
    URF_CHI_TITLE: "U-Chi",
    URF_CTS_TITLE: "CTS",
    URF_DDP_TITLE: "DDP",
    URF_IT_TITLE: "IT",
    URF_CIT_TITLE: "CIT",
    XLEARNER_TITLE: "XL"
}

############################## FOLDERS ##############################
RESULS = "/results/"
METRICS = RESULS + "metrics/"
FIGURES = RESULS + "figures/"



import logging
import pickle
from datetime import datetime

import numpy as np
from causalml.inference.meta import BaseXClassifier
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from autoum.approaches.utils import ApproachParameters, DataSetsHelper, Helper


# TODO: Cost sensitive learning
# TODO: Feature Importance


class XLearner:
    """
    X-Learner proposed by Künzel et al. (2019)

    X-learner is an extension of T-learner (Two-Model Approach), and consists of three stages as follows:

    Stage 1
    Estimate the response functions using any machine learning algorithm

    𝜇1(𝑥)=𝐸[𝑌(1)|𝑋=𝑥]; 𝜇'1=> Response function for the treatment group "treatment estimator", and
    𝜇0(𝑥)=𝐸[𝑌(0)|𝑋=𝑥]; 𝜇'0=> Response function for the control group "control estimator"

    Stage 2
    Impute the individiual treatment effects:

    𝐷1𝑖=𝑌1𝑖−𝜇̂ 0(𝑋1𝑖); ITE for the treatment individuals is defined as difference between true outcome of the treatment individual minus the outcome of the control estimator, and
    𝐷0𝑖=𝜇̂ 1(𝑋0𝑖)−𝑌0𝑖; ITE for the control individuals is defined as difference between the outcome of the treatment estimator and the true outcome of the control individual

    Example:
        Individuum A was in the treatment group and responded (that is what we have observed):
            * True Outcome: Treatment Responder
            * Outcome of the control estimator: e.g., Control Non-Responder
            * ITE (𝐷1): Treatment Responder - Control Non-Responder (e.g., 1-0 = 1)

    After imputing the ITE, estimate 𝜏1(𝑥)=𝐸[𝐷1|𝑋=𝑥], and 𝜏0(𝑥)=𝐸[𝐷0|𝑋=𝑥] using any machine learning algorithm.
    That is, we will use the imputed treatment effects D1 and D0 as response variable to obtain 𝜏1 and 𝜏0 (CATE)

    Stage 3
    Define the CATE estimate by a weighted average of 𝜏1(𝑥) and 𝜏0(𝑥):

    𝜏(𝑥)=𝑔(𝑥)𝜏0(𝑥)+(1−𝑔(𝑥))𝜏1(𝑥)

    where 𝑔∈[0,1] is a weight funciton. We can use propensity scores for 𝑔(𝑥).
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for the  X-learner approach

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """
        self.parameters = parameters
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using the X-Learner.

        Side note: For randomized controlled experiments, we can use the "observed propensity score", that is, the ratio between the individuals who were subject to a treatment
        and the individuals who were not subject to a treatment. For observational data, we need to estimate the propensity score.

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        # Add causalML's 'treatment' column
        experiment_groups_col = Helper.add_treatment_group_key(data_set_helper.df_train)

        # Initiate outcome and effect learner
        outcome_learner = RandomForestClassifier(**self.parameters)  # treatment and control estimator - stage 1. Should be a classifier
        effect_learner = RandomForestRegressor(**self.parameters)  # effect (CATE, tau) estimator - stage 2. Should be a regressor

        self.log.debug("Start fitting X-Learner ...")

        # Get the covariates for the data set
        x_train = data_set_helper.x_train
        x_test = data_set_helper.x_test

        # X Learner
        num_treatment_samples = np.count_nonzero(data_set_helper.df_train['treatment'].to_numpy() == 1)
        x_learner = BaseXClassifier(outcome_learner=outcome_learner, effect_learner=effect_learner, control_name='c')
        x_learner.fit(X=x_train, treatment=experiment_groups_col, y=data_set_helper.y_train, p=Helper.get_propensity_score(x_train.shape[0], num_treatment_samples))

        self.log.debug(x_learner)

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_X-Learner_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(x_learner, open(filename, 'wb'))

        self.log.debug("Predicting ...")

        score_train = x_learner.predict(X=x_train, p=Helper.get_propensity_score(x_train.shape[0]))
        score_test = x_learner.predict(X=x_test, p=Helper.get_propensity_score(x_test.shape[0]))

        if data_set_helper.valid:
            x_valid = data_set_helper.x_valid
            score_valid = x_learner.predict(X=x_valid, p=Helper.get_propensity_score(x_valid.shape[0]))
        else:
            score_valid = np.array([0], np.int32)

        return {
            "score_train": score_train.flatten(),
            "score_valid": score_valid.flatten(),
            "score_test": score_test.flatten()
        }



import logging
import time
import warnings

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

warnings.filterwarnings("ignore", category=FutureWarning)

pd.options.mode.chained_assignment = None
pd.set_option('display.max_columns', 10)

from autoum.pipelines.pipeline_rw import PipelineRW
from autoum.const.const import *


class PipelineSD(PipelineRW):
    """
    Pipeline with synthetic data sets

    The synthetic data is created based on the method suggest by Wager & Athey, 2018, Estimation and Inference of Heterogeneous Treatment Effects using Random Forests, p.1238

    Either use k-fold Cross Validation or regular training/test split
    """

    def __init__(self, n: int, p: int, sigma: float, threshold: float, propensity: float, **kwargs):
        # Inheritance from parent class (PipelineRW)
        super().__init__(**kwargs)

        self.n = n
        self.p = p
        self.sigma = sigma
        self.threshold = threshold
        self.propensity = propensity

    def analyze_dataset(self, data: pd.DataFrame = None):
        """
        Apply, compare, and evaluate various uplift modeling approaches on synthetic data

        :param data: Dataset to be analyzed. Default: None as we want to analyze a synthetic dataset
        """
        logging.info(f"run_name: {self.run_name}, run_id: {self.run_name}, sigma: {self.sigma}, p: {self.p}, threshold: {self.threshold}, propensity: {self.propensity}")

        # Set random seed
        np.random.seed(self.random_seed)

        # Set start time
        start = time.time()

        # Create synthetic dataset
        df = self.create_synthetic_dataset()

        df_train, df_test = train_test_split(df, test_size=self.test_size, shuffle=True, stratify=df[['response', 'treatment']], random_state=self.random_seed)
        df_train.reset_index(inplace=True, drop=True)
        df_test.reset_index(inplace=True, drop=True)

        feature_names = list(df.drop(['response', 'treatment'], axis=1).columns.values)

        # If self.max_features is set to auto, set the number of features for the direct uplift approach to sqrt(n_features)
        if self.max_features == 'auto':
            n_features = len(feature_names)
            self.parameters[URF_TITLE + '_parameters']['max_features'] = int(round(np.sqrt(n_features)))

        # k-fold cross validation
        if self.cv_number_splits > 2:
            dict_uplift = self.analyze_k_fold_cv(df_train=df_train, df_test=df_test)

        # regular train/test split
        else:
            dict_uplift = self.analyze_single_fold(df_train=df_train, df_test=df_test)

        # Calculate, plot and save metrics
        self.calculate_metrics(list_feature_importances=dict_uplift["feature_importances"], feature_names=feature_names,
                               list_dict_uplift_train=dict_uplift["list_dict_uplift_train"], list_dict_uplift_valid=dict_uplift["list_dict_uplift_valid"],
                               list_dict_uplift_test=dict_uplift["list_dict_uplift_test"], list_dict_opt_uplift_train=dict_uplift["list_dict_opt_uplift_train"],
                               list_dict_opt_uplift_valid=dict_uplift["list_dict_opt_uplift_valid"], list_dict_opt_uplift_test=dict_uplift["list_dict_opt_uplift_test"])

        end = time.time()
        logging.info('Function took %0.3f s' % (end - start))

    def create_synthetic_dataset(self):
        """
        Create a complex, non-linear CATE function

        :return: DataFrame (synthetic data)
        """
        means = np.zeros(self.p)
        cov_mat = np.ones([self.p, self.p]) * 0.3 - (np.eye(self.p) * (0.3 - 1))
        x = np.random.multivariate_normal(means, cov_mat, size=self.n)
        e = np.repeat(self.propensity, self.n)
        w = np.random.binomial(1, e, size=self.n).reshape((self.n, -1))
        x = np.append(x, w, axis=1)

        def rho(x):
            return 2 / (1 + np.exp(-12 * (x - (1 / 2))))

        mu_1 = 1 / 2 * rho(x[:, 0]) * rho(x[:, 1])
        mu_0 = -1 / 2 * rho(x[:, 0]) * rho(x[:, 1])
        y_1 = mu_1 + self.sigma * np.random.normal(size=self.n)
        y_0 = mu_0 + self.sigma * np.random.normal(size=self.n)
        y = np.where(x[:, -1] == 1, y_1, y_0)
        y_binary = np.where(y > np.percentile(y, self.threshold), 1, 0)
        x = x[:, :-1]

        columns = [f"Feature{i}" for i in range(0, x.shape[1])]
        df = pd.DataFrame(data=x, columns=columns)
        df["response"] = y_binary
        df["treatment"] = w

        return df



import codecs
import os
from subprocess import check_call
from sys import platform

from setuptools import find_packages, setup
from setuptools.command.install import install

root = os.getcwd()

packages = find_packages(exclude=["tests", "tests.*"])

with open("requirements.txt") as f:
    requirements = f.readlines()

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        check_call("git clone https://github.com/jroessler/causalml.git".split(), cwd=root)
        check_call("python setup.py build_ext --inplace".split(), cwd=root + "/causalml")
        if platform == "linux" or platform == "linux2":
            check_call("python setup.py install".split(), cwd=root + "/causalml")
        else:
            check_call("pip install --no-cache-dir .".split(), cwd=root + "/causalml")
        check_call("pip install xbcausalforest==0.1.3".split())


setup(name='autoum',
      version=get_version("autoum/__init__.py"),
      description='A Python Framework for Automatically Evaluating various Uplift Modeling Algorithms to Estimate Individual Treatment Effects',
      url='https://github.com/jroessler/autoum',
      author='Jannik Rößler',
      author_email="",
      packages=packages,
      cmdclass={'install': PostInstallCommand},
      python_requires=">=3.8",
      install_requires=requirements,
      classifiers=["Programming Language :: Python", "License :: OSI Approved :: Apache Software License", "Operating System :: OS Independent"]
      )


import itertools as it

import numpy as np
import pandas as pd
from scipy.stats import f, rankdata, studentized_range


class Statistics:
    """
    Class for statistical comparison of classifiers over multiple data sets

    Documentation only! Please aware that we need a scipy version > 1.4.1 but causaml and econml require a causaml version == 1.4.1
    """

    @staticmethod
    def calculate_ranks(a, y_col, group_col, block_col, melted, sort):
        """
        Given an np.ndarray with size (k, n), calculate the ranks for each data set and algorithm

        See Statistics.nemenyi_srs for parameter description
        """
        from scikit_posthocs import __convert_to_block_df

        x, _y_col, _group_col, _block_col = __convert_to_block_df(a, y_col, group_col, block_col, melted)
        x = x.sort_values(by=[_group_col, _block_col], ascending=True) if sort else x
        x.dropna(inplace=True)

        groups = x[_group_col].unique()
        k = groups.size
        n = x[_block_col].unique().size

        # Calculate ranks
        x['mat'] = x.groupby(_block_col)[_y_col].rank()
        R = x.groupby(_group_col)['mat'].mean()
        vs = np.zeros((k, k))
        combs = it.combinations(range(k), 2)

        return combs, R, x, groups, vs, k, n

    @staticmethod
    def nemenyi_srs(a, alpha, columns, y_col=None, block_col=None, group_col=None, melted=False, sort=False):
        """
        Post-hoc Nemenyi test with Studentized Rank Statistics (own implementation):

        The performance of two classifiers i and j is significant different if the corresponding average ranks R_i and R_j differ by
        at least the critical difference:

        CD = q_a*\sqrt{k(k+1)/6N}

        where critical values q_a are based on the Studentized range statistic divided by \sqrt{2}

        :param a: array_like or pandas DataFrame object. An array, any object exposing the array interface or a pandas DataFrame.
        :param alpha: Alpha value
        :param columns: Name of the columns
        :param y_col: Must be specified if `a` is a pandas DataFrame object. Name of the column that contains y data.
        :param block_col: Must be specified if `a` is a pandas DataFrame object. Name of the column that contains blocking factor values.
        :param group_col: Must be specified if `a` is a pandas DataFrame object. Name of the column that contains treatment (group) factor values.
        :param melted: Specifies if data are given as melted columns "y", "blocks", and "groups". Default: False
        :param sort: If True, sort data by block and group columns. Default: False
        :return: DataFrame containing 0 and 1, whereby 1 refers to a pairwise comparison with a significant difference with alpha

        """
        combs, R, x, groups, vs, k, n = Statistics.calculate_ranks(a, y_col, group_col, block_col, melted, sort)

        tri_upper = np.triu_indices(vs.shape[0], 1)
        tri_lower = np.tril_indices(vs.shape[0], -1)
        vs[:, :] = 0
        significances = np.full((vs.shape[0], vs.shape[1]), True)

        # Calculate pairwise difference in average ranks over the data sets
        for i, j in combs:
            vs[i, j] = np.abs(R[groups[i]] - R[groups[j]])

        # Get the critical value
        cv = studentized_range.ppf(1 - alpha, k, np.inf) / np.sqrt(2)

        # Calculate the critical difference
        cd = cv * np.sqrt((k * (k + 1)) / (6 * n))

        # Return 1 if the difference in ranks is equal or higher than the critical difference
        significances[tri_upper] = vs[tri_upper] >= cd
        significances[tri_lower] = np.transpose(vs)[tri_lower] >= cd
        significances = significances.astype(int)
        np.fill_diagonal(significances, 0)

        df = pd.DataFrame(significances, index=groups, columns=groups)

        df.set_axis(columns, inplace=True)
        df.set_axis(columns, axis=1, inplace=True)

        return df

    @staticmethod
    def calculate_friedman_test(*args):
        """
        Friedman test (non-parametric) with N (# of data sets) and k (# of algorithms) are big (e.g., N>10 and k>5). In our case it is usally k>10 and N>5.

        Null hypothesis: There are no differences in performance

        1. Calculate the ranks for each data set (rank the algorithm per data sets)
        2. Calculate the chi-squared approximation of the Friedman test
        3. Calculate the number of degrees of freedom (k-1)
        4. Calculate the F-Distribution
        4. Check the Table of critical value for the F distribution using a significance value alpha and the degrees of freedom
        5. Reject the null hypothesis if the calculated F-Distribution is equal to or greater than the table critical chi-squared value at the prespecified level of significance

        :param args: (list of array, each containing the performance on an algorithm over the n data sets)
        :return: The p-value
        """
        k = len(args)
        if k < 3:
            raise ValueError('Less than 3 levels.  Friedman test not appropriate.')

        n = len(args[0])
        for i in range(1, k):
            if len(args[i]) != n:
                raise ValueError('Unequal N in friedmanchisquare.  Aborting.')

        # Rank data
        data = np.vstack(args).T
        data = data.astype(float)
        for i in range(len(data)):
            data[i] = k + 1 - rankdata(data[i])

        # Calculate the sum of average ranks of all algorithms
        ssbn = np.sum((data.sum(axis=0) / n) ** 2)

        # Approximate the chi-sqaure of the Friedman test
        chisq = ((12.0 * n) / (k * (k + 1))) * (ssbn - ((k * (k + 1) ** 2) / 4))

        # Calculate the F-distribution
        f_dist = (n - 1) * chisq / (n * (k - 1) - chisq)

        # Calculate the p-value
        p = f.sf(f_dist, k - 1, (k - 1) * (n - 1))

        return round(p, 4)


import logging
import pickle
from datetime import datetime

import numpy as np
from xbcausalforest import XBCF

from autoum.approaches.utils import ApproachParameters, DataSetsHelper


# TODO: Cost sensitive learning
# TODO: Feature Importance


class BayesianCausalForest:
    """
    Bayesian Causal Forest (BCF) algorithm proposed by Hahn et al. (2020)

    This algorithm is built on Bayesian additive regression tree (BART)
    An important insight is that the specification of the causal model is a sum of two parts:
    E[Y=y | X=x, T=t]= μ(x) + τ(x)t

    The first part μ(x) is a BART model to estimate the expected outcome from a set of covariates and, if required, an estimate of the probability to receive treatment.
    The parameters for this BART model are denoted with pr for ‘prognostic’, e.g. num_trees_pr.
    The second part τ(x) is a BART model that estimates the treatment effect conditional on some covariates, with its parameters denoted by trt as in ‘treatment’,
    e.g., num_trees_trt
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters):
        """
        Creates a classifier for the Bayesian Causal Forest (Hahn et al, 2020)

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        """

        self.parameters = parameters
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using the Bayesian causal forest.

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        # Prior on the variance in the leaf for each tree. Hahn et al. propose to scale the prior with some factor of the variance of the outcome divided by the number of trees.
        # Default values from here: https://johaupt.github.io/bart/bayesian/causal%20inference/xbcf.html
        self.parameters["tau_pr"] = 0.6 * np.var(data_set_helper.y_train) / self.parameters["num_trees_pr"]
        self.parameters["tau_trt"] = 0.1 * np.var(data_set_helper.y_train) / self.parameters["num_trees_trt"]

        # Count categorical columns
        categorical_cols = 0
        for col in data_set_helper.df_train.drop(["treatment", "response"], axis=1).columns:
            if len(data_set_helper.df_train[col].value_counts()) <= 2:
                categorical_cols += 1
        self.parameters["p_categorical_pr"] = categorical_cols
        self.parameters["p_categorical_trt"] = categorical_cols

        bcf_classifier = XBCF(**self.parameters)

        self.log.debug("Start fitting BayesianCausalForest ...")

        bcf_classifier.fit(x_t=data_set_helper.x_train, x=data_set_helper.x_train, y=data_set_helper.y_train, z=data_set_helper.df_train.treatment)

        self.log.debug(bcf_classifier)

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_BCF_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(bcf_classifier, open(filename, 'wb'))

        self.log.debug("Predicting ... ")

        # Note: The .predict() method returns a ndarray with the treatment effect for each sample (the higher the value, the more likely the individual is to accept if treated)
        score_train = bcf_classifier.predict(data_set_helper.x_train)
        score_test = bcf_classifier.predict(data_set_helper.x_test)
        if data_set_helper.valid:
            score_valid = bcf_classifier.predict(data_set_helper.x_valid)
        else:
            score_valid = np.array([0], np.int32)

        return {
            "score_train": score_train,
            "score_valid": score_valid,
            "score_test": score_test
        }



import logging
import pickle
from datetime import datetime

import numpy as np
from causalml.inference.tree import UpliftRandomForestClassifier

from autoum.approaches.utils import ApproachParameters, DataSetsHelper, Helper


# TODO: Cost sensitive learning
# TODO: Feature Importance


class UpliftRandomForest:
    """
    Uplift Random Forest proposed by Rzepakowski and Jaroszewicz (2012)

    Modify existing supervised learning algorithms to directly infer a causal effect.
    According to the current literature, decision trees and different ensembles of decision trees are the most popular adjusted algorithms.
    Usually, the tree-building algorithm and the splitting criterion are modified such that they maximize the difference in uplift.

    The Uplift Tree approach consists of a set of methods that use a tree-based algorithm (Random Forest) where the
    splitting criterion is based on differences in uplift (N. Rzepakowski, Jaroszewicz, 2012) and (Hansotia & Rukstales, 2002)

    You can choose among the following evaluation functions:
    "ED": Euclidean Distance
    "KL": Kullback-Leibler Divergence
    "Chi" Chi-Square Divergence
    "DDP": Delta Delta P
    "CTS": Contextual Treatment Selection
    "IT": Interaction Tree (Su et al., 2009)
    "CIT": Causal Inference Tree (Su et al., 2012)
    """

    def __init__(self, parameters: dict, approach_parameters: ApproachParameters, eval_function: str):
        """
        Creates a classifier for the uplift random forest

        :param parameters: The parameters needed for the creation of the base learner
        :param approach_parameters: Pass an approach_parameters object that contains all parameters necessary to execute the approach
        :param eval_function: Evaluation function. You can choose among the following evaluation functions:
            "ED": Euclidean Distance
            "KL": Kullback-Leibler Divergence
            "Chi" Chi-Square Divergence
            "DDP": Delta Delta P
            "CTS": Contextual Treatment Selection
            "IT": Interaction Tree (Su et al., 2009)
            "CIT": Causal Inference Tree (Su et al., 2012)
        """

        self.parameters = parameters
        self.parameters["evaluationFunction"] = eval_function
        self.feature_importance = approach_parameters.feature_importance
        self.save = approach_parameters.save
        self.path = approach_parameters.path
        self.split_number = approach_parameters.split_number
        self.log = logging.getLogger(type(self).__name__)

    def analyze(self, data_set_helper: DataSetsHelper) -> dict:
        """
        Calculate the score (ITE/Uplift/CATE) for each sample using uplift random forest

        :param data_set_helper: A DataSetsHelper comprising the training, validation (optional) and test data set
        :return: Dictionary containing, scores and feature importance
        """

        # Add causalML's 'treatment' column
        experiment_groups_col = Helper.add_treatment_group_key(data_set_helper.df_train)

        # Initiate UpliftRandomForestClassifier
        urf = UpliftRandomForestClassifier(**self.parameters)

        self.log.debug("Start fitting Uplift Random Forest ...")

        urf.fit(X=data_set_helper.x_train, treatment=experiment_groups_col, y=data_set_helper.y_train)

        self.log.debug(urf)

        if self.save:
            self.log.debug("Saving ...")
            date_str = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
            filename = self.path + 'results/models/{}_Direct_Uplift_{}.pickle'.format(str(self.split_number), date_str)
            pickle.dump(urf, open(filename, 'wb'))

        self.log.debug("Predicting ... ")

        # Note: The .predict() method returns an ndarray in which each column contains the predicted uplift if the unit was in the corresponding treatment group.
        score_train = urf.predict(data_set_helper.x_train)
        score_test = urf.predict(data_set_helper.x_test)
        if data_set_helper.valid:
            score_valid = urf.predict(data_set_helper.x_valid)
        else:
            score_valid = np.array([0], np.int32)

        return {
            "score_train": score_train.flatten(),
            "score_valid": score_valid.flatten(),
            "score_test": score_test.flatten()
        }
